"""Widgets for the submission of bands work chains.

Authors: AiiDAlab team
"""

import ipywidgets as ipw

from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.panel import ConfigurationSettingsPanel
from aiidalab_qe.common.widgets import HBoxWithUnits

from .hubbard import (
    HubbardConfigurationSettingsModel,
    HubbardConfigurationSettingsPanel,
)
from .magnetization import (
    MagnetizationConfigurationSettingsModel,
    MagnetizationConfigurationSettingsPanel,
)
from .model import AdvancedConfigurationSettingsModel
from .pseudos import (
    PseudosConfigurationSettingsModel,
    PseudosConfigurationSettingsPanel,
)
from .smearing import (
    SmearingConfigurationSettingsModel,
    SmearingConfigurationSettingsPanel,
)


class AdvancedConfigurationSettingsPanel(
    ConfigurationSettingsPanel[AdvancedConfigurationSettingsModel],
):
    def __init__(self, model: AdvancedConfigurationSettingsModel, **kwargs):
        super().__init__(
            model=model,
            layout={"justify_content": "space-between", **kwargs.get("layout", {})},
            **kwargs,
        )

        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )
        self._model.observe(
            self._on_protocol_change,
            "protocol",
        )
        self._model.observe(
            self._on_kpoints_distance_change,
            "kpoints_distance",
        )

        # NOTE connect pseudos first, as some settings depend on it
        pseudos_model = PseudosConfigurationSettingsModel()
        self.pseudos = PseudosConfigurationSettingsPanel(model=pseudos_model)
        model.add_model("pseudos", pseudos_model)

        smearing_model = SmearingConfigurationSettingsModel()
        self.smearing = SmearingConfigurationSettingsPanel(model=smearing_model)
        model.add_model("smearing", smearing_model)

        magnetization_model = MagnetizationConfigurationSettingsModel()
        self.magnetization = MagnetizationConfigurationSettingsPanel(
            model=magnetization_model,
        )
        model.add_model("magnetization", magnetization_model)

        hubbard_model = HubbardConfigurationSettingsModel()
        self.hubbard = HubbardConfigurationSettingsPanel(model=hubbard_model)
        model.add_model("hubbard", hubbard_model)

    def render(self):
        if self.rendered:
            return

        self.reset_to_defaults_button = ipw.Button(
            description="Reset to defaults",
            button_style="primary",
            icon="undo",
            layout=ipw.Layout(width="fit-content"),
        )
        self.reset_to_defaults_button.on_click(self._on_reset_to_defaults_button_click)

        self.clean_workdir = ipw.Checkbox(
            description="Delete the work directory after the calculation",
            indent=False,
            layout=ipw.Layout(width="fit-content", margin="5px 2px"),
        )
        ipw.link(
            (self._model, "clean_workdir"),
            (self.clean_workdir, "value"),
        )

        # Smearing setting widget
        self.smearing.render()

        # Kpoints setting widget
        self.kpoints_distance = ipw.BoundedFloatText(
            min=0.0,
            step=0.05,
            description="K-points distance:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "kpoints_distance"),
            (self.kpoints_distance, "value"),
        )
        ipw.dlink(
            (self._model, "input_structure"),
            (self.kpoints_distance, "disabled"),
            lambda _: not self._model.has_pbc,
        )
        self.mesh_grid = ipw.HTML(layout=ipw.Layout(margin="0 0 0 10px"))
        ipw.dlink(
            (self._model, "mesh_grid"),
            (self.mesh_grid, "value"),
        )

        # Hubbard setting widget
        self.hubbard.render()

        # Total change setting widget
        self.total_charge = ipw.BoundedFloatText(
            min=-3,
            max=3,
            step=0.01,
            description="Total charge:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "total_charge"),
            (self.total_charge, "value"),
        )

        # Van der Waals setting widget
        self.van_der_waals = ipw.Dropdown(
            description="Van der Waals correction:",
            style={"description_width": "150px"},
        )
        ipw.dlink(
            (self._model, "van_der_waals_options"),
            (self.van_der_waals, "options"),
        )
        ipw.link(
            (self._model, "van_der_waals"),
            (self.van_der_waals, "value"),
        )

        # Magnetization settings
        self.magnetization.render()

        # Convergence Threshold settings
        self.scf_conv_thr = ipw.BoundedFloatText(
            min=1e-15,
            max=1.0,
            description="SCF:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "scf_conv_thr"),
            (self.scf_conv_thr, "value"),
        )
        ipw.dlink(
            (self._model, "scf_conv_thr_step"),
            (self.scf_conv_thr, "step"),
        )
        self.forc_conv_thr = ipw.BoundedFloatText(
            min=1e-15,
            max=1.0,
            format="0.0e",
            description="Force:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "forc_conv_thr"),
            (self.forc_conv_thr, "value"),
        )
        ipw.dlink(
            (self._model, "forc_conv_thr_step"),
            (self.forc_conv_thr, "step"),
        )
        self.etot_conv_thr = ipw.BoundedFloatText(
            min=1e-15,
            max=1.0,
            format="0.0e",
            description="Energy:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "etot_conv_thr"),
            (self.etot_conv_thr, "value"),
        )
        ipw.dlink(
            (self._model, "etot_conv_thr_step"),
            (self.etot_conv_thr, "step"),
        )
        self.electron_maxstep = ipw.BoundedIntText(
            min=20,
            max=1000,
            step=1,
            description="Electronic:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "electron_maxstep"),
            (self.electron_maxstep, "value"),
        )

        self.optimization_maxsteps = ipw.BoundedIntText(
            min=50,
            max=1000,
            step=1,
            description="Ionic:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "optimization_maxsteps"),
            (self.optimization_maxsteps, "value"),
        )
        self.pseudos.render()

        self.children = [
            InAppGuide(identifier="advanced-settings"),
            self.reset_to_defaults_button,
            self.clean_workdir,
            self.total_charge,
            self.van_der_waals,
            ipw.HTML("<h2>Convergence</h2>"),
            ipw.HTML("""
                <div style="line-height: 1.4; margin-bottom: 5px;">
                    Control the convergence criteria of the self-consistent field (SCF)
                    geometry optimization cycles.
                </div>
            """),
            ipw.HTML("<h4>Thresholds</h4>"),
            ipw.HTML("""
                <div style="line-height: 1.4; margin-bottom: 5px;">
                    Setting thresholds for energy, force, and self-consistency ensures calculation accuracy and stability.
                    <br>
                    Lower values increase the accuracy but also the computational cost.
                    <br>
                    The default values are set by the <b>protocol</b> are usually a
                    good starting point.
                </div>
            """),
            HBoxWithUnits(self.forc_conv_thr, "Ry/Bohr"),
            HBoxWithUnits(self.etot_conv_thr, "Ry/atom"),
            HBoxWithUnits(self.scf_conv_thr, "Ry/atom"),
            ipw.HTML("<h4>Maximum cycle steps</h4>"),
            ipw.HTML("""
                <div style="line-height: 1.4; margin-bottom: 5px;">
                    Setting a maximum number of electronic and ionic optimization steps
                    ensures that the calculation does not run indefinitely.
                </div>
            """),
            self.electron_maxstep,
            self.optimization_maxsteps,
            self.smearing,
            ipw.HTML("<h2>K-points</h2>"),
            ipw.HTML("""
                <div style="line-height: 1.4; margin-bottom: 5px;">
                    The k-points mesh density of the SCF calculation is set by the
                    <b>protocol</b>.
                    <br>
                    The value below represents the maximum distance between k-points
                    in each direction of reciprocal space.
                    <br>
                    Smaller is more accurate and costly.
                </div>
            """),
            ipw.HBox(
                children=[
                    HBoxWithUnits(self.kpoints_distance, "Å<sup>-1</sup>"),
                    self.mesh_grid,
                ],
                layout=ipw.Layout(align_items="center"),
            ),
            self.magnetization,
            self.hubbard,
            self.pseudos,
        ]

        self.rendered = True

        self.refresh()

    def _on_input_structure_change(self, _):
        self.refresh(specific="structure")

    def _on_protocol_change(self, _):
        self.refresh(specific="protocol")

    def _on_kpoints_distance_change(self, _):
        self.refresh(specific="mesh")

    def _on_reset_to_defaults_button_click(self, _):
        self._reset()

    def _reset(self):
        self._model.reset()
        for _, model in self._model.get_models():
            model.reset()



from aiidalab_qe.app.submission import SubmitQeAppWorkChainStep
from aiidalab_qe.app.submission.model import SubmissionStepModel
from aiidalab_qe.app.wizard_app import WizardApp


def test_code_not_selected(submit_app_generator):
    """Test if there is an error when the code is not selected."""
    app: WizardApp = submit_app_generator(properties=["dos"])
    model = app.submit_model
    model.get_model("global").get_model("quantumespresso__dos").selected = None
    # Check builder construction passes without an error
    parameters = model.get_model_state()
    model._create_builder(parameters)


def test_set_selected_codes(submit_app_generator):
    """Test set_selected_codes method."""
    app: WizardApp = submit_app_generator()
    parameters = app.submit_model.get_model_state()
    model = SubmissionStepModel()
    _ = SubmitQeAppWorkChainStep(model=model, auto_setup=False)
    for identifier, code_model in app.submit_model.get_model("global").get_models():
        model.get_model("global").get_model(identifier).is_active = code_model.is_active
    model.qe_installed = True
    model.get_model("global").set_selected_codes(parameters["codes"]["global"]["codes"])
    assert model.get_selected_codes() == app.submit_model.get_selected_codes()


def test_update_codes_display(app: WizardApp):
    """Test update_codes_display method.
    If the workchain property is not selected, the related code should be hidden.
    """
    app.submit_step.render()
    model = app.submit_model
    global_model = model.get_model("global")
    global_model.update_active_codes()
    global_resources = app.submit_step.global_resources
    assert global_resources.code_widgets["dos"].layout.display == "none"
    model.input_parameters = {"workchain": {"properties": ["pdos"]}}
    global_model.update_active_codes()
    assert global_model.get_model("quantumespresso__dos").is_active is True
    assert global_resources.code_widgets["dos"].layout.display == "block"


def test_check_blockers(app: WizardApp):
    """Test check_submission_blockers method."""
    model = app.submit_model

    model.update_blockers()
    assert len(model.blockers) == 0

    model.input_parameters = {"workchain": {"properties": ["pdos"]}}
    model.update_blockers()
    assert len(model.blockers) == 0

    # set dos code to None, will introduce another blocker
    dos_code = model.get_model("global").get_model("quantumespresso__dos")
    dos_value = dos_code.selected
    dos_code.selected = None
    model.update_blockers()
    assert len(model.blockers) == 1

    # set dos code back will remove the blocker
    dos_code.selected = dos_value
    model.update_blockers()
    assert len(model.blockers) == 0


def test_qeapp_computational_resources_widget(app: WizardApp):
    """Test QEAppComputationalResourcesWidget."""
    app.submit_step.render()
    global_model = app.submit_model.get_model("global")
    global_resources = app.submit_step.global_resources
    pw_code_model = global_model.get_model("quantumespresso__pw")
    pw_code_widget = global_resources.code_widgets["pw"]
    assert pw_code_widget.parallelization.npool.layout.display == "none"
    pw_code_model.parallelization_override = True
    pw_code_model.npool = 2
    assert pw_code_widget.parallelization.npool.layout.display == "block"
    assert pw_code_widget.parameters == {
        "code": global_resources.code_widgets["pw"].value,
        "cpus": 1,
        "cpus_per_task": 1,
        "max_wallclock_seconds": 43200,
        "nodes": 1,
        "ntasks_per_node": 1,
        "parallelization": {"npool": 2},
    }



from __future__ import annotations

import threading
import typing as t
from copy import deepcopy

import ipywidgets as ipw
import traitlets as tl

from aiida import orm
from aiida.common.links import LinkType
from aiida.engine import ProcessState
from aiida.tools.graph.graph_traversers import traverse_graph
from aiidalab_qe.common.mixins import HasProcess
from aiidalab_qe.common.mvc import Model
from aiidalab_qe.common.widgets import LoadingWidget

from .state import STATE_ICONS


class SimplifiedProcessTreeModel(Model, HasProcess):
    clicked = tl.Unicode(None, allow_none=True)


class SimplifiedProcessTree(ipw.VBox):
    def __init__(self, model: SimplifiedProcessTreeModel, **kwargs):
        self.loading_message = LoadingWidget("Loading process tree")
        super().__init__(
            children=[self.loading_message],
            **kwargs,
        )
        self.add_class("simplified-process-tree")
        self._model = model
        self._model.observe(
            self._on_process_change,
            names="process_uuid",
        )
        self._model.observe(
            self._on_monitor_counter_change,
            "monitor_counter",
        )
        self.rendered = False

    def render(self):
        if self.rendered:
            return
        if self._model.has_process:
            self._render()
        else:
            self.children = [ipw.HTML("Process node not yet available")]

    def _on_process_change(self, _):
        if not self.rendered:
            self._render()
        self._update()

    def _on_monitor_counter_change(self, _):
        if self.rendered:
            self._update()

    def _on_inspect(self, uuid: str):
        self._model.clicked = None  # ensure event is triggered when label is reclicked
        self._model.clicked = uuid

    def _render(self):
        self.collapse_button = ipw.Button(
            description="Collapse all",
            button_style="warning",
            icon="minus",
            tooltip="Collapse all branches",
            layout=ipw.Layout(
                width="fit-content",
                margin="0 0 10px 0",
            ),
        )
        self.collapse_button.on_click(self._collapse_all)

        root = self._model.fetch_process_node()

        self.trunk = WorkChainTreeNode(node=root, on_inspect=self._on_inspect)
        self.trunk.add_class("tree-trunk")
        self.trunk.initialize()
        self.trunk.expand()
        self._update()

        self.tree_container = ipw.VBox(
            children=[self.trunk],
        )
        self.tree_container.add_class("tree-container")

        self.children = [
            self.collapse_button,
            self.tree_container,
            ipw.HBox(
                children=[
                    ipw.HTML(
                        value="""
                        <div style="font-style: italic;">
                            *Actual number of jobs may exceed estimated total due to
                            error handling, dynamic sub-workflows, and/or other runtime
                            adjustments
                        </div>
                    """,
                    ),
                ],
                layout=ipw.Layout(
                    align_items="flex-start",
                    margin="5px 10px 0 0",
                ),
            ),
        ]

        self.rendered = True

    def _update(self):
        self.trunk.update()

    def _collapse_all(self, _):
        self.trunk.collapse(recursive=True)


ProcessNodeType = t.TypeVar("ProcessNodeType", bound=orm.ProcessNode)


class ProcessTreeNode(ipw.VBox, t.Generic[ProcessNodeType]):
    _TITLE_MAPPING = {
        "QeAppWorkChain": "Quantum ESPRESSO app workflow",
        "BandsWorkChain": "Electronic band structure workflow",
        "PwBandsWorkChain": "Electronic band structure workflow",
        "ProjwfcBandsWorkChain": "Electronic band structure workflow",
        "PwRelaxWorkChain": "Structure relaxation workflow manager",
        "PdosWorkChain": "Projected density of states workflow",
        "PwBaseWorkChain": {
            "scf": "SCF workflow",
            "nscf": "NSCF workflow",
            "bands": "Bands workflow",
            "relax": "Structure relaxation workflow",
            "md": "Molecular dynamics workflow",
        },
        "PwCalculation": {
            "scf": "Run SCF cycle",
            "nscf": "Run NSCF cycle",
            "bands": "Compute bands",
            "relax": "Optimize structure geometry",
            "md": "Run molecular dynamics simulation",
        },
        "DosCalculation": "Compute density of states",
        "ProjwfcCalculation": "Compute projections",
    }

    def __init__(
        self,
        node: ProcessNodeType,
        level: int = 0,
        on_inspect: t.Callable[[str], None] | None = None,
        **kwargs,
    ):
        self.uuid = node.uuid
        self.level = level
        self.on_inspect = on_inspect
        super().__init__(**kwargs)
        self._node: dict[int, ProcessNodeType] = {}  # thread_id: node

    @property
    def node(self) -> ProcessNodeType:
        if (tid := threading.get_ident()) not in self._node:
            node = orm.load_node(self.uuid)
            self._node[tid] = node
        return self._node[tid]

    def initialize(self):
        self._build_header()
        self.children = [self.header]

    def update(self):
        self.state.value = self._get_state()
        self.emoji.value = self._get_emoji(self.state.value)

    def _build_header(self):
        self.title = self._get_human_readable_title()
        self.indentation = self._get_indentation()
        self.emoji = ipw.HTML()
        self.state = ipw.HTML()
        self.header = ipw.HBox()
        self.header.add_class("tree-node-header")

    def _get_indentation(self):
        return ipw.HTML(layout=ipw.Layout(width=f"{22 * self.level}px"))

    def _get_emoji(self, state):
        return STATE_ICONS.get(state, "❓")

    def _get_state(self):
        if not hasattr(self.node, "process_state"):
            return "queued"
        if self.node.is_failed:
            return "failed"
        state = self.node.process_state
        return (
            "running"
            if state is ProcessState.WAITING
            else state.value
            if state
            else "created"
        )

    def _get_human_readable_title(self):
        node = self.node
        if not (label := node.process_label):
            return "Unknown"
        if label in ("PwBaseWorkChain", "PwCalculation"):
            inputs = node.inputs.pw if label == "PwBaseWorkChain" else node.inputs
            calculation: str = inputs.parameters.get_dict()["CONTROL"]["calculation"]
            calculation = calculation.replace("vc-", "")
            return self._TITLE_MAPPING.get(label, {}).get(calculation, label)
        return self._TITLE_MAPPING.get(label, label)


class ProcessTreeBranches(ipw.VBox):
    def clear(self):
        self.children = []

    def __len__(self):
        return len(self.children)

    def __getitem__(self, index: int) -> ProcessTreeNode:
        return self.children[index]  # type: ignore

    def __iter__(self) -> t.Iterator[ProcessTreeNode]:
        return iter(self.children)  # type: ignore

    def __iadd__(self, other: ProcessTreeNode):
        if not isinstance(other, ProcessTreeNode):
            raise TypeError("Right operand must be a TreeNode")
        self.children += (other,)
        return self


class WorkChainTreeNode(ProcessTreeNode[orm.WorkChainNode]):
    _adding_branches = False

    @property
    def metadata_inputs(self):
        # BACKWARDS COMPATIBILITY: originally this was added because "relax" was in the
        # metadata inputs regardless if running relaxation, as it was used also for the
        # summary to extract pw parameters. #1163 removes this dependency, thus allowing
        # the popping of the "relax" port if not running relaxation. However, this is
        # kept for backwards compatibility, for jobs ran before #1163.
        inputs = deepcopy(self.node.get_metadata_inputs()) or {}
        if "properties" in self.node.inputs:
            inputs = {
                key: value
                for key, value in inputs.items()
                if key in self.node.inputs.properties
            }
        return inputs

    @property
    def sub_processes(self):
        return set(self.metadata_inputs.keys()) - {"metadata"}

    @property
    def has_sub_workflows(self):
        return len(self.sub_processes) > 1

    @property
    def collapsed(self):
        return self.toggle.icon == "plus"

    def initialize(self):
        super().initialize()
        self.pks = set()
        self.branches = ProcessTreeBranches()
        self.branches.add_class("tree-node-branches")
        self.children += (self.branches,)
        self.expected_jobs = self._get_expected(self.metadata_inputs)
        self.update()

    def update(self):
        super().update()
        self.tally.value = self._get_tally()
        if not self.collapsed:
            self._add_branches()
        for branch in self.branches:
            branch.update()

    def clear(self):
        self.branches.clear()
        self.pks.clear()

    def expand(self, recursive=False):
        if self.collapsed:
            self.toggle.click()
        if recursive:
            for branch in self.branches:
                if isinstance(branch, WorkChainTreeNode):
                    branch.expand(recursive=True)

    def collapse(self, recursive: bool = False):
        if not self.collapsed:
            self.toggle.click()
        if recursive:
            for branch in self.branches:
                if isinstance(branch, WorkChainTreeNode):
                    branch.collapse(recursive=True)

    def _build_header(self):
        super()._build_header()
        self.toggle = ipw.Button(icon="plus")
        self.toggle.add_class("tree-node-toggle")
        self.toggle.on_click(self._toggle_branches)
        self.label = ipw.HTML(self.title)
        self.tally = ipw.HTML()
        self.header.children = [
            self.indentation,
            self.toggle,
            self.emoji,
            self.label,
            ipw.HTML(" | "),
            self.state,
            ipw.HTML(" | "),
            self.tally,
        ]

    def _add_branches(self):
        if self._adding_branches:
            return
        self._adding_branches = True
        self._add_branches_recursive()
        self._adding_branches = False

    def _add_branches_recursive(self, node: orm.ProcessNode | None = None):
        node = node or self.node
        for child in sorted(node.called, key=lambda child: child.ctime):
            if isinstance(child, orm.CalcFunctionNode):
                continue
            if child.pk in self.pks:
                continue
            if child.process_label in (
                "BandsWorkChain",
                "ProjwfcBaseWorkChain",
            ):
                self._add_branches_recursive(child)
            else:
                self._add_branch(child)

    def _add_branch(self, child: orm.ProcessNode):
        TreeNodeClass = (
            WorkChainTreeNode
            if isinstance(child, orm.WorkChainNode)
            else CalcJobTreeNode
        )
        branch = TreeNodeClass(
            node=child,
            level=self.level + 1,
            on_inspect=self.on_inspect,
        )
        branch.initialize()
        self.branches += branch
        self.pks.add(child.pk)

    def _get_tally(self):
        total = self.expected_jobs["count"]
        dynamic = self.expected_jobs["dynamic"]
        finished = self._count_finished()
        tally = f"{finished}/{total}"
        tally += "*" if dynamic else ""
        tally += " job" if total == 1 else " jobs"
        return tally

    def _count_finished(self):
        traverser = traverse_graph(
            starting_pks=[self.node.pk],
            links_forward=[
                LinkType.CALL_WORK,
                LinkType.CALL_CALC,
            ],
        )
        return (
            orm.QueryBuilder()
            .append(
                orm.CalcJobNode,
                filters=(
                    (orm.CalcJobNode.fields.pk.in_(traverser.get("nodes")))
                    & (orm.CalcJobNode.fields.process_state == "finished")
                    & (orm.CalcJobNode.fields.exit_status == 0)
                ),
            )
            .count()
        )

    def _get_expected(self, inputs: dict[str, dict]) -> dict:
        expected = {}
        count = 0
        dynamic = False

        for key, sub_inputs in inputs.items():
            if "metadata" in sub_inputs and "options" in sub_inputs["metadata"]:
                # This is a calculation
                count += 1
                dynamic = dynamic or key == "pw"
            elif key != "metadata":
                # This is a workflow
                nested = self._get_expected(sub_inputs)
                count += nested["count"]
                dynamic = dynamic or nested["dynamic"]
                expected[key] = nested

        expected |= {
            "count": count,
            "dynamic": dynamic,
        }

        return expected

    def _toggle_branches(self, _):
        if self.collapsed:
            self.branches.add_class("open")
            self.toggle.icon = "minus"
            self._add_branches()
        else:
            self.branches.remove_class("open")
            self.toggle.icon = "plus"


class CalcJobTreeNode(ProcessTreeNode[orm.CalcJobNode]):
    def initialize(self):
        super().initialize()
        self.update()

    def _build_header(self):
        super()._build_header()
        self.label = ipw.Button(
            description=self.title,
            tooltip="click to view calculation",
        )
        self.label.add_class("calculation-link")
        self.label.on_click(self._on_label_click)
        self.header.children = [
            self.indentation,
            self.emoji,
            self.label,
            ipw.HTML(" | "),
            self.state,
        ]

    def _on_label_click(self, _):
        if self.on_inspect is None:
            return
        self.on_inspect(self.uuid)



def test_result(generate_qeapp_workchain):
    import plotly.graph_objects as go

    from aiidalab_qe.common.bands_pdos import BandsPdosWidget
    from aiidalab_qe.plugins.electronic_structure.result import (
        ElectronicStructureResultsModel,
        ElectronicStructureResultsPanel,
    )

    workchain = generate_qeapp_workchain(run_bands=False)
    model = ElectronicStructureResultsModel()
    panel = ElectronicStructureResultsPanel(model=model)
    model.process_uuid = workchain.node.uuid

    assert model.title == "Electronic PDOS"
    assert model.identifiers == ["pdos"]

    panel.render()

    assert len(panel.results_container.children) == 1  # only pdos, so no controls

    widget = panel.bands_pdos_container.children[0]  # type: ignore
    model = widget._model

    assert isinstance(widget, BandsPdosWidget)
    assert isinstance(widget.plot, go.FigureWidget)

    assert not model.bands_data
    assert model.pdos_data

    assert widget.plot.layout.xaxis.title.text == "Density of states (eV)"
    assert widget.plot.layout.yaxis.title.text is None



import ipywidgets as ipw
import traitlets as tl

from aiida.engine import ProcessState
from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.process import STATE_ICONS
from aiidalab_qe.common.widgets import LoadingWidget
from aiidalab_qe.common.wizard import QeDependentWizardStep
from aiidalab_widgets_base import ProcessMonitor, WizardAppWidgetStep

from .components import ResultsComponent
from .components.status import WorkChainStatusModel, WorkChainStatusPanel
from .components.summary import WorkChainSummary, WorkChainSummaryModel
from .components.viewer import WorkChainResultsViewer, WorkChainResultsViewerModel
from .model import ResultsStepModel


class ViewQeAppWorkChainStatusAndResultsStep(QeDependentWizardStep[ResultsStepModel]):
    missing_information_warning = (
        "No available results. Did you submit or load a calculation?"
    )

    STATUS_TEMPLATE = "<h4>Workflow status: {}</h4"

    def __init__(self, model: ResultsStepModel, **kwargs):
        self.log_widget = kwargs.pop("log_widget", None)
        super().__init__(model=model, **kwargs)
        self.observe(
            self._on_previous_step_state_change,
            "previous_step_state",
        )
        self._model.observe(
            self._on_process_change,
            "process_uuid",
        )

    def _render(self):
        self.kill_button = ipw.Button(
            description="Kill workchain",
            tooltip="Kill the below workchain.",
            button_style="danger",
            icon="stop",
            layout=ipw.Layout(width="auto", display="none"),
        )
        ipw.dlink(
            (self, "state"),
            (self.kill_button, "disabled"),
            lambda state: state is not self.State.ACTIVE,
        )
        self.kill_button.on_click(self._on_kill_button_click)

        self.clean_scratch_button = ipw.Button(
            description="Clean remote data",
            tooltip="Clean the remote folders of the workchain.",
            button_style="danger",
            icon="trash",
            layout=ipw.Layout(width="auto", display="none"),
        )
        ipw.dlink(
            (self._model, "process_remote_folder_is_clean"),
            (self.clean_scratch_button, "disabled"),
        )
        self.clean_scratch_button.on_click(self._on_clean_scratch_button_click)

        self.process_info = ipw.HTML()
        ipw.dlink(
            (self._model, "process_info"),
            (self.process_info, "value"),
        )

        summary_model = WorkChainSummaryModel()
        self.summary_panel = WorkChainSummary(model=summary_model)
        self._model.add_model("summary", summary_model)

        results_model = WorkChainResultsViewerModel()
        self.results_panel = WorkChainResultsViewer(model=results_model)
        self._model.add_model("results", results_model)

        status_model = WorkChainStatusModel()
        self.status_panel = WorkChainStatusPanel(model=status_model)
        self._model.add_model("status", status_model)

        self.panels = {
            "Summary": self.summary_panel,
            "Status": self.status_panel,
            "Results": self.results_panel,
        }

        self.toggle_controls = ipw.ToggleButtons(
            options=[*self.panels.keys()],
            tooltips=[
                "A summary of calculation parameters",
                "The calculation results",
                "A detailed progress status of the workflow",
            ],
            icons=[
                "file-text-o",
                "bar-chart",
                "tasks",
            ],
            value=None,
        )
        self.toggle_controls.add_class("results-step-toggles")
        self.toggle_controls.observe(
            self._on_toggle_change,
            "value",
        )

        self.container = ipw.VBox(
            children=[
                self.status_panel,
            ],
        )

        if self._model.has_process:
            self._update_children()

    def _post_render(self):
        self._update_kill_button_layout()
        self._update_clean_scratch_button_layout()

        self.toggle_controls.value = (
            "Results"
            if (process := self._model.fetch_process_node()) and process.is_finished_ok
            else "Status"
        )

        self.process_monitor = ProcessMonitor(
            timeout=0.5,
            callbacks=[
                self._update_status,
                self._update_state,
            ],
            log_widget=self.log_widget,
        )
        ipw.dlink(
            (self._model, "process_uuid"),
            (self.process_monitor, "value"),
        )

    def can_reset(self):
        "Checks if process is running (active), which disallows a reset."
        return self.state is not self.State.ACTIVE

    def reset(self):
        self._model.reset()

    @tl.observe("state")
    def _on_state_change(self, change):
        super()._on_state_change(change)
        self._update_kill_button_layout()

    def _on_previous_step_state_change(self, _):
        if self.previous_step_state is WizardAppWidgetStep.State.SUCCESS:
            process_node = self._model.fetch_process_node()
            message = (
                "Loading results"
                if process_node and process_node.is_finished
                else "Submitting calculation"
            )
            self.children = [LoadingWidget(message)]

    def _on_toggle_change(self, change):
        panel = self.panels[change["new"]]
        self._toggle_view(panel)

    def _on_process_change(self, _):
        if self.rendered:
            self._update_children()
        self._model.update()
        self._update_state()
        self._update_kill_button_layout()
        self._update_clean_scratch_button_layout()

    def _on_kill_button_click(self, _):
        self._model.kill_process()
        self._update_kill_button_layout()

    def _on_clean_scratch_button_click(self, _):
        self._model.clean_remote_data()
        self._update_clean_scratch_button_layout()

    def _update_children(self):
        self.children = [
            InAppGuide(identifier="results-step"),
            self.process_info,
            ipw.HBox(
                children=[
                    self.kill_button,
                    self.clean_scratch_button,
                ],
                layout=ipw.Layout(margin="0 3px"),
            ),
            self.toggle_controls,
            self.container,
        ]
        self.previous_children = list(self.children)

    def _toggle_view(self, panel: ResultsComponent):
        self.container.children = [panel]
        panel.render()

    def _update_kill_button_layout(self):
        if not self.rendered:
            return
        process_node = self._model.fetch_process_node()
        if (
            not process_node
            or process_node.is_finished
            or process_node.is_excepted
            or self.state
            in (
                self.State.SUCCESS,
                self.State.FAIL,
            )
        ):
            self.kill_button.layout.display = "none"
        else:
            self.kill_button.layout.display = "block"

    def _update_clean_scratch_button_layout(self):
        if not self.rendered:
            return
        process_node = self._model.fetch_process_node()
        if process_node and process_node.is_terminated:
            self.clean_scratch_button.layout.display = "block"
        else:
            self.clean_scratch_button.layout.display = "none"

    def _update_status(self):
        self._model.monitor_counter += 1

    def _update_state(self):
        if not (process_node := self._model.fetch_process_node()):
            self.state = self.State.INIT
            self._update_controls()
            return

        if process_state := process_node.process_state:
            status = self._get_process_status(process_state.value)
        else:
            status = "Unknown"

        if process_state is ProcessState.CREATED:
            self.state = self.State.ACTIVE
        elif process_state in (
            ProcessState.RUNNING,
            ProcessState.WAITING,
        ):
            self.state = self.State.ACTIVE
            status = self._get_process_status("running")  # overwrite status
        elif process_state in (
            ProcessState.EXCEPTED,
            ProcessState.KILLED,
        ):
            self.state = self.State.FAIL
        elif process_node.is_failed:
            self.state = self.State.FAIL
        elif process_node.is_finished_ok:
            self.state = self.State.SUCCESS

        self._model.process_info = self.STATUS_TEMPLATE.format(status)

        self._update_controls()

    def _update_controls(self):
        if self.state in (self.State.SUCCESS, self.State.FAIL):
            self._update_kill_button_layout()
            self._update_clean_scratch_button_layout()

    def _get_process_status(self, state: str):
        return f"{state.capitalize()} {STATE_ICONS[state]}"



from collections.abc import Mapping
from importlib import resources
from pathlib import Path

import yaml

from aiidalab_qe.app import parameters


def recursive_merge(d1, d2):
    """Merge dictionary `d2` into dictionary `d1` recursively.

    - For keys that are in both dictionaries:
        - If values are dictionaries, merge them recursively.
        - Otherwise, overwrite the value in `d1` with the value in `d2`.
    - Keys in `d2` not in `d1` are added to `d1`.

    Parameters
    ----------
    `d1` : `dict`
        The dictionary to merge into.
    `d2` : `dict`
        The dictionary to merge from.

    Returns
    -------
    `dict`
        The merged dictionary.

    Examples
    --------
    >>> d1 = {'a': 1, 'b': {'c': 2, 'd': 3}}
    >>> d2 = {'b': {'c': 4, 'e': 5}}
    >>> recursive_merge(d1, d2)
    {'a': 1, 'b': {'c': 4, 'd': 3, 'e': 5}}
    """
    for key, value in d2.items():
        if key in d1:
            if isinstance(d1[key], Mapping) and isinstance(value, Mapping):
                recursive_merge(d1[key], value)
            else:
                d1[key] = value
        else:
            d1[key] = value
    return d1


DEFAULT_PARAMETERS = yaml.safe_load(resources.read_text(parameters, "qeapp.yaml"))

CONFIG_DIR_PATH = Path.home() / ".aiidalab" / "quantum-espresso"

custom_config_file = CONFIG_DIR_PATH / "qe-config.yml"
if custom_config_file.exists():
    custom_config = yaml.safe_load(custom_config_file.read_text())
    DEFAULT_PARAMETERS = recursive_merge(DEFAULT_PARAMETERS, custom_config)



import ipywidgets as ipw
import traitlets as tl

from aiida import orm
from aiida.common import NotExistent
from aiidalab_qe.common.mvc import Model
from aiidalab_qe.common.widgets import (
    PwCodeResourceSetupWidget,
    QEAppComputationalResourcesWidget,
)


class CodeModel(Model):
    is_active = tl.Bool(False)
    options = tl.List(
        trait=tl.Tuple(tl.Unicode(), tl.Unicode()),  # code option (label, uuid)
        default_value=[],
    )
    selected = tl.Unicode(default_value=None, allow_none=True)
    num_nodes = tl.Int(1)
    num_cpus = tl.Int(1)
    ntasks_per_node = tl.Int(1)
    cpus_per_task = tl.Int(1)
    max_wallclock_seconds = tl.Int(3600 * 12)
    allow_hidden_codes = tl.Bool(False)
    allow_disabled_computers = tl.Bool(False)
    override = tl.Bool(False)

    def __init__(
        self,
        *,
        name="",
        description,
        default_calc_job_plugin,
        code_widget_class=QEAppComputationalResourcesWidget,
    ):
        self.name = name
        self.description = description
        self.default_calc_job_plugin = default_calc_job_plugin
        self.code_widget_class = code_widget_class
        self.is_rendered = False

        ipw.dlink(
            (self, "num_cpus"),
            (self, "ntasks_per_node"),
        )

    @property
    def is_ready(self):
        return self.is_active and bool(self.selected)

    @property
    def first_option(self):
        return self.options[0][1] if self.options else None  # type: ignore

    def activate(self):
        self.is_active = True

    def deactivate(self):
        self.is_active = False

    def update(self, user_email="", refresh=False):
        if not self.options or refresh:
            self.options = self._get_codes(user_email)
            self.selected = self.first_option

    def get_model_state(self) -> dict:
        return {
            "options": self.options,
            "code": self.selected,
            "nodes": self.num_nodes,
            "cpus": self.num_cpus,
            "ntasks_per_node": self.ntasks_per_node,
            "cpus_per_task": self.cpus_per_task,
            "max_wallclock_seconds": self.max_wallclock_seconds,
        }

    def set_model_state(self, parameters: dict):
        self.selected = (
            self._get_uuid(identifier)
            if (identifier := parameters.get("code"))
            else self.first_option
        )
        self.num_nodes = parameters.get("nodes", 1)
        self.num_cpus = parameters.get("cpus", 1)
        self.ntasks_per_node = parameters.get("ntasks_per_node", 1)
        self.cpus_per_task = parameters.get("cpus_per_task", 1)
        self.max_wallclock_seconds = parameters.get("max_wallclock_seconds", 3600 * 12)

    def _get_uuid(self, identifier):
        try:
            uuid = orm.load_code(identifier).uuid
        except NotExistent:
            uuid = None
        # If the code was imported from another user, it is not usable
        # in the app and thus will not be considered as an option!
        return uuid if uuid in [opt[1] for opt in self.options] else None

    def _get_codes(self, user_email: str = ""):
        user = orm.User.collection.get(email=user_email)

        filters = (
            {"attributes.input_plugin": self.default_calc_job_plugin}
            if self.default_calc_job_plugin
            else {}
        )

        codes = (
            orm.QueryBuilder()
            .append(
                orm.Code,
                filters=filters,
            )
            .all(flat=True)
        )

        return [
            (self._full_code_label(code), code.uuid)
            for code in codes
            if code.computer.is_user_configured(user)
            and (self.allow_hidden_codes or not code.is_hidden)
            and (self.allow_disabled_computers or code.computer.is_user_enabled(user))
        ]

    @staticmethod
    def _full_code_label(code):
        return f"{code.label}@{code.computer.label}"


class PwCodeModel(CodeModel):
    parallelization_override = tl.Bool(False)
    npool = tl.Int(1)

    def __init__(
        self,
        *,
        name="",
        description="pw.x",
        default_calc_job_plugin="quantumespresso.pw",
        code_widget_class=PwCodeResourceSetupWidget,
    ):
        super().__init__(
            name=name,
            description=description,
            default_calc_job_plugin=default_calc_job_plugin,
            code_widget_class=code_widget_class,
        )

    def get_model_state(self) -> dict:
        parameters = super().get_model_state()
        parameters["parallelization"] = (
            {
                "npool": self.npool,
            }
            if self.parallelization_override
            else {}
        )
        return parameters

    def set_model_state(self, parameters):
        super().set_model_state(parameters)
        if "parallelization" in parameters and "npool" in parameters["parallelization"]:
            self.parallelization_override = True
            self.npool = parameters["parallelization"].get("npool", 1)
        else:
            self.parallelization_override = False


CodesDict = dict[str, CodeModel]
PluginCodes = dict[str, CodesDict]



"""Widgets for the QE app.

Authors: AiiDAlab team
"""

import base64
import hashlib
import subprocess
import typing as t
from copy import deepcopy
from queue import Queue
from tempfile import NamedTemporaryFile
from threading import Event, Lock, Thread
from time import time

import anywidget
import ase
import ipywidgets as ipw
import numpy as np
import requests as req
import traitlets
from IPython.display import HTML, Javascript, clear_output, display
from pymatgen.io.ase import AseAtomsAdaptor
from shakenbreak.distortions import distort, local_mc_rattle, rattle

from aiida.orm import CalcJobNode, load_code, load_node
from aiida.orm import Data as orm_Data
from aiidalab_widgets_base import (
    ComputationalResourcesWidget,
    StructureExamplesWidget,
)
from aiidalab_widgets_base.utils import (
    StatusHTML,
    list_to_string_range,
    string_range_to_list,
)

__all__ = [
    "CalcJobOutputFollower",
    "LogOutputWidget",
]


class RollingOutput(ipw.VBox):
    style = (
        "background-color: #253239; color: #cdd3df; line-height: normal; custom=test"
    )

    value = traitlets.Unicode()
    auto_scroll = traitlets.Bool()

    def __init__(self, num_min_lines=10, max_output_height="200px", **kwargs):  # noqa: ARG002
        self._num_min_lines = num_min_lines
        self._output = ipw.HTML(layout=ipw.Layout(min_width="80em"))
        self._refresh_output()
        super().__init__(
            children=[
                self._output,
            ],
            layout=ipw.Layout(max_height=max_output_height),
        )
        self.add_class("rolling-output")

    @traitlets.default("value")
    def _default_value(self):
        if self._num_min_lines > 0:
            return "\n" * self._num_min_lines

    @traitlets.default("auto_scroll")
    def _default_auto_scroll(self):
        return True

    def scroll_to_bottom(self):
        # Slight hack because it will scroll all widgets with the same class
        # name and max height to the bottom. That would primarily be an issue in
        # case that there are two LogOutputWidgets in the DOM. Could probably be
        # alleviated by adding a custom class.
        display(
            Javascript(
                """
            Array.from(document.getElementsByClassName('{class_name}'))
            .filter(el => el.style["max-height"] === "{max_height}")
            .forEach(el => el.scrollTop = el.scrollHeight)""".format(
                    class_name="p-Widget p-Panel jupyter-widgets widget-container widget-box widget-vbox",
                    max_height=self.layout.max_height,
                )
            )
        )

    @traitlets.observe("value")
    def _refresh_output(self, _=None):
        self._output.value = self._format_output(self.value)
        if self.auto_scroll:
            self.scroll_to_bottom()

    def _format_output(self, text):
        lines = text.splitlines()

        # Add empty lines to reach the minimum number of lines.
        lines += [""] * max(0, self._num_min_lines - len(lines))

        # Replace empty lines with single white space to ensure that they are
        # actually shown.
        lines = [line if len(line) > 0 else " " for line in lines]

        text = "\n".join(lines)
        return f"""<pre style="{self.style}">{text}</pre>"""


class DownloadButton(ipw.Button):
    # Adapted from https://stackoverflow.com/a/68683463
    """A Button widget for downloads with dynamic content."""

    filename = traitlets.Unicode()
    payload = traitlets.Bytes()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.on_click(self.__on_click)

    @traitlets.default("icon")
    def _default_icon(self):
        return "download"

    @traitlets.default("tooltip")
    def _default_tooltip(self):
        return "Download"

    def __on_click(self, _):
        digest = hashlib.md5(self.payload).hexdigest()  # bypass browser cache
        payload = base64.b64encode(self.payload).decode()

        link_id = f"dl_{digest}"

        display(
            HTML(
                f"""
            <html>
            <body>
            <a id="{link_id}" download="{self.filename}" href="data:text/plain;base64,{payload}" download>
            </a>

            <script>
            (function download() {{
            document.getElementById('{id}').click();
            }})()
            </script>

            </body>
            </html>
            """
            )
        )


class FilenameDisplayWidget(ipw.Box):
    value = traitlets.Unicode()

    def __init__(self, max_width=None, **kwargs):
        self.max_width = max_width
        self._html = ipw.HTML()
        super().__init__([self._html], **kwargs)
        self.add_class("filename-display")

    @traitlets.observe("value")
    def _observe_filename(self, change):
        icon = '<i class="fa fa-file-text-o" aria-hidden="true"></i>'
        width_style = f"width:{self.max_width};" if self.max_width else ""
        self._html.value = f"""
        <div style="
            white-space:nowrap;
            overflow:hidden;
            text-overflow:ellipsis;
            {width_style}">
            {icon} {change["new"]}
        </div>
        """


class LogOutputWidget(ipw.VBox):
    filename = traitlets.Unicode()
    value = traitlets.Unicode()

    def __init__(self, placeholder=None, **kwargs):
        self.placeholder = placeholder

        self._rolling_output = RollingOutput(
            layout=ipw.Layout(flex="1 1 auto"),
            max_output_height="unset",
        )
        ipw.dlink(
            (self, "value"),
            (self._rolling_output, "value"),
            lambda value: value or self.placeholder or "",
        )

        self._filename_display = FilenameDisplayWidget(layout=ipw.Layout(width="auto"))
        ipw.dlink(
            (self, "filename"),
            (self._filename_display, "value"),
            lambda value: value or "[no filename]",
        )

        self._btn_download = DownloadButton(
            layout=ipw.Layout(width="30px", flex="5 1 auto"),
            disabled=True,
        )
        ipw.dlink((self, "filename"), (self._btn_download, "filename"))
        ipw.dlink(
            (self, "value"),
            (self._btn_download, "payload"),
            transform=lambda value: value.encode("utf-8"),
        )

        self._btn_scroll_down = ipw.Button(
            icon="angle-double-down",
            tooltip="Scroll to bottom",
            layout=ipw.Layout(width="30px", flex="1 1 auto"),
            disabled=True,
        )
        self._btn_scroll_down.on_click(
            lambda _: self._rolling_output.scroll_to_bottom()
        )

        self._btn_auto_scroll = ipw.ToggleButton(
            icon="magic",
            tooltip="Auto-scroll with content",
            value=True,
            layout=ipw.Layout(flex="1 1 auto", width="30px"),
        )
        ipw.link(
            (self._btn_auto_scroll, "value"), (self._rolling_output, "auto_scroll")
        )

        self._btns = ipw.VBox(
            [self._btn_download, self._btn_scroll_down, self._btn_auto_scroll],
            layout=ipw.Layout(min_width="36px", flex_flow="column"),
        )

        super().__init__(
            children=[
                self._filename_display,
                ipw.HBox(
                    children=[
                        self._rolling_output,
                        self._btns,
                    ],
                    layout=ipw.Layout(height="100%"),
                ),
            ],
            layout=ipw.Layout(height="100%"),
            **kwargs,
        )
        self.add_class("log-output")

    @traitlets.default("placeholder")
    def _default_placeholder(self):
        return "[empty]"

    @traitlets.observe("value")
    def _observe_value(self, change):
        self._btn_download.disabled = not change["new"]
        self._btn_scroll_down.disabled = not change["new"]


class CalcJobOutputFollower(traitlets.HasTraits):
    calcjob_uuid = traitlets.Unicode(allow_none=True)
    filename = traitlets.Unicode(allow_none=True)
    output = traitlets.List(trait=traitlets.Unicode)
    lineno = traitlets.Int()

    def __init__(self, **kwargs):
        self._output_queue = Queue()

        self._lock = Lock()
        self._push_thread = None
        self._pull_thread = None
        self._stop_follow_output = Event()
        self._follow_output_thread = None

        super().__init__(**kwargs)

    @traitlets.observe("calcjob_uuid")
    def _observe_calcjob(self, change):
        calcjob_uuid = change["new"]
        if change["old"] == calcjob_uuid:
            return

        with self._lock:
            # Stop following
            self._stop_follow_output.set()

            if self._follow_output_thread:
                self._follow_output_thread.join()
                self._follow_output_thread = None

            # Reset all traitlets and signals.
            self.output.clear()
            self.lineno = 0
            self._stop_follow_output.clear()

            # (Re/)start following
            if change["new"]:
                self._follow_output_thread = Thread(
                    target=self._follow_output, args=(calcjob_uuid,)
                )
                self._follow_output_thread.start()

    def _follow_output(self, calcjob_uuid):
        """Monitor calcjob and orchestrate pushing and pulling of output."""
        self._pull_thread = Thread(target=self._pull_output)
        self._pull_thread.start()
        self._push_thread = Thread(target=self._push_output, args=(calcjob_uuid,))
        self._push_thread.start()

    def _fetch_output(self, calcjob):
        assert isinstance(calcjob, CalcJobNode)
        if "retrieved" in calcjob.outputs:
            try:
                self.filename = calcjob.base.attributes.get("output_filename")
                with calcjob.outputs.retrieved.base.repository.open(self.filename) as f:
                    return f.read().splitlines()
            except OSError:
                return []

        elif "remote_folder" in calcjob.outputs:
            try:
                fn_out = calcjob.base.attributes.get("output_filename")
                self.filename = fn_out
                with NamedTemporaryFile() as tmpfile:
                    calcjob.outputs.remote_folder.getfile(fn_out, tmpfile.name)
                    return tmpfile.read().decode().splitlines()
            except OSError:
                return []
        else:
            return []

    _EOF = None

    def _push_output(self, calcjob_uuid, delay=0.2):
        """Push new log lines onto the queue."""
        lineno = 0
        calcjob = load_node(calcjob_uuid)
        while True:
            try:
                lines = self._fetch_output(calcjob)
            except Exception as error:
                self._output_queue.put([f"[ERROR: {error}]"])
            else:
                self._output_queue.put(lines[lineno:])
                lineno = len(lines)
            finally:
                if calcjob.is_sealed or self._stop_follow_output.wait(delay):
                    # Pushing EOF signals to the pull thread to stop.
                    self._output_queue.put(self._EOF)
                    break  # noqa: B012

    def _pull_output(self):
        """Pull new log lines from the queue and update traitlets."""
        while True:
            item = self._output_queue.get()
            if item is self._EOF:
                self._output_queue.task_done()
                break
            else:  # item is 'new lines'
                with self.hold_trait_notifications():
                    self.output.extend(item)
                    self.lineno += len(item)
                self._output_queue.task_done()


class ProgressBar(ipw.HBox):
    class AnimationRate(float):
        pass

    description = traitlets.Unicode()
    value = traitlets.Union([traitlets.Float(), traitlets.Instance(AnimationRate)])
    bar_style = traitlets.Unicode()

    _animation_rate = traitlets.Float()

    def __init__(self, description_layout=None, *args, **kwargs):
        if description_layout is None:
            description_layout = ipw.Layout(width="auto", flex="2 1 auto")

        self._label = ipw.Label(layout=description_layout)
        self._progress_bar = ipw.FloatProgress(
            min=0, max=1.0, layout=ipw.Layout(width="auto", flex="1 1 auto")
        )

        traitlets.link((self, "description"), (self._label, "value"))
        traitlets.link((self, "bar_style"), (self._progress_bar, "bar_style"))

        self._animate_stop_event = Event()
        self._animate_thread = None

        super().__init__([self._label, self._progress_bar], *args, **kwargs)

    def _animate(self, refresh_rate=0.01):
        v0 = self._progress_bar.value
        t0 = time()

        while not self._animate_stop_event.wait(refresh_rate):
            self._progress_bar.value = (v0 + (time() - t0) * self._animation_rate) % 1.0

    def _start_animate(self):
        if self._animate_thread is not None:
            raise RuntimeError("Cannot start animation more than once!")

        self._animate_thread = Thread(target=self._animate)
        self._animate_thread.start()

    def _stop_animate(self):
        self._animate_stop_event.set()
        self._animate_thread.join()
        self._animate_stop_event.clear()
        self._animate_thread = None

    @traitlets.default("_animation_rate")
    def _default_animation_rate(self):
        return 0

    @traitlets.observe("_animation_rate")
    def _observe_animation_rate(self, change):
        if change["new"] and not change["old"]:
            self._start_animate()
        elif not change["new"] and change["old"]:
            self._stop_animate()

    @traitlets.validate("value")
    def _validate_value(self, proposal):
        if isinstance(proposal["value"], self.AnimationRate):
            if proposal["value"] < 0:
                raise traitlets.TraitError("The animation rate must be non-negative.")

        elif not 0 <= proposal["value"] <= 1.0:
            raise traitlets.TraitError("The value must be between 0 and 1.0.")

        return proposal["value"]

    @traitlets.observe("value")
    def _observe_value(self, change):
        if isinstance(change["new"], self.AnimationRate):
            self._animation_rate = change["new"]
        else:
            self._animation_rate = 0
            self._progress_bar.value = change["new"]


class AddingTagsEditor(ipw.VBox):
    """Editor for adding tags to atoms."""

    structure = traitlets.Instance(ase.Atoms, allow_none=True)
    selection = traitlets.List(traitlets.Int, allow_none=True)
    input_selection = traitlets.List(traitlets.Int, allow_none=True)
    structure_node = traitlets.Instance(orm_Data, allow_none=True, read_only=True)

    def __init__(self, title="", **kwargs):
        self.title = title

        self._status_message = StatusHTML()
        self.atom_selection = ipw.Text(
            placeholder="e.g. 1..5 8 10",
            description="Index of atoms",
            value="",
            style={"description_width": "100px"},
            layout={"width": "initial"},
        )
        self.from_selection = ipw.Button(description="From selection")
        self.from_selection.on_click(self._from_selection)
        self.tag = ipw.BoundedIntText(
            description="Tag", value=1, min=0, max=11, layout={"width": "initial"}
        )
        self.add_tags = ipw.Button(
            description="Update tags",
            button_style="primary",
            layout={"width": "initial"},
        )

        self.reset_tags = ipw.Button(
            description="Reset tags",
            button_style="primary",
            layout={"width": "initial"},
        )
        self.reset_all_tags = ipw.Button(
            description="Reset all tags",
            button_style="warning",
            layout={"width": "initial"},
        )
        self.scroll_note = ipw.HTML(
            value="<p style='font-style: italic;'>Note: The table is scrollable.</p>",
            layout={"visibility": "hidden"},
        )
        self.tag_display = ipw.Output()
        self.add_tags.on_click(self._add_tags)
        self.reset_tags.on_click(self._reset_tags)
        self.reset_all_tags.on_click(self._reset_all_tags)
        self.atom_selection.observe(self._display_table, "value")
        self.add_tags.on_click(self._display_table)
        self.reset_tags.on_click(self._display_table)
        self.reset_all_tags.on_click(self._display_table)

        super().__init__(
            children=[
                ipw.HTML(
                    """
                    <p>
                    These are used to distinguish atoms of the same chemical element. <br>
                    For example, they can be used to assign different initial magnetization values for antiferromagnetic systems.
                    </p>
                    <p style="font-weight: bold; color: #1f77b4;">NOTE:</p>
                    <ul style="padding-left: 2em; list-style-type: disc;">
                        <li>Atom indices start from 1, not 0. This means that the first atom in the list is numbered 1, the second atom is numbered 2, and so on.</li>
                    </ul>
                    </p>
                    """
                ),
                ipw.HBox(
                    [
                        self.atom_selection,
                        self.from_selection,
                        self.tag,
                    ]
                ),
                self.tag_display,
                self.scroll_note,
                ipw.HBox([self.add_tags, self.reset_tags, self.reset_all_tags]),
                self._status_message,
            ],
            **kwargs,
        )

    def _display_table(self, _=None):
        """Function to control tag_display
        When given a list of atom in selection it will display a HTML table with Index, Element and Tag
        """
        selection = string_range_to_list(self.atom_selection.value)[0]
        current_tags = self.structure.get_tags()
        chemichal_symbols = self.structure.get_chemical_symbols()

        if (
            selection
            and (min(selection) >= 0)
            and (max(selection) <= (len(self.structure) - 1))
        ):
            table_data = []
            for index in selection:
                tag = current_tags[index]
                symbol = chemichal_symbols[index]
                if tag == 0:
                    tag = ""
                table_data.append([f"{index + 1}", f"{symbol}", f"{tag}"])

            # Create an HTML table
            table_html = "<table>"
            table_html += "<tr><th>Index</th><th>Element</th><th>Tag</th></tr>"
            for row in table_data:
                table_html += "<tr>"
                for cell in row:
                    table_html += f"<td>{cell}</td>"
                table_html += "</tr>"
            table_html += "</table>"

            # Set layout to a fix size
            self.tag_display.layout = {
                "overflow": "auto",
                "height": "100px",
                "width": "150px",
            }
            with self.tag_display:
                clear_output()
                display(HTML(table_html))
            self.scroll_note.layout = {"visibility": "visible"}
        else:
            self.tag_display.layout = {}
            with self.tag_display:
                clear_output()
            self.scroll_note.layout = {"visibility": "hidden"}

    def _from_selection(self, _=None):
        """Set the atom selection from the current selection."""
        self.atom_selection.value = list_to_string_range(self.selection)

    def _add_tags(self, _=None):
        """Add tags to the selected atoms."""
        if not self.atom_selection.value:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select atoms first.</strong>
            </div>
            """
        else:
            selection = string_range_to_list(self.atom_selection.value)[0]
            new_structure = deepcopy(self.structure)
            if not new_structure.get_tags().any():
                new_tags = np.zeros(len(new_structure))
            else:
                new_tags = new_structure.get_tags()
            new_tags[selection] = self.tag.value
            new_structure.set_tags(new_tags)
            self.structure = None
            self.structure = deepcopy(new_structure)
            self.input_selection = None
            self.input_selection = deepcopy(self.selection)

    def _reset_tags(self, _=None):
        """Clear tags from selected atoms."""
        if not self.atom_selection.value:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select atoms first.</strong>
            </div>
            """
        else:
            selection = string_range_to_list(self.atom_selection.value)[0]
            new_structure = deepcopy(self.structure)
            new_tags = new_structure.get_tags()
            new_tags[selection] = 0
            new_structure.set_tags(new_tags)
            self.structure = None
            self.structure = deepcopy(new_structure)
            self.input_selection = None
            self.input_selection = deepcopy(self.selection)

    def _reset_all_tags(self, _=None):
        """Clear all tags."""
        new_structure = deepcopy(self.structure)
        new_tags = np.zeros(len(new_structure))
        new_structure.set_tags(new_tags)
        self.structure = None
        self.structure = deepcopy(new_structure)
        self.input_selection = None
        self.input_selection = deepcopy(self.selection)


class PeriodicityEditor(ipw.VBox):
    """Editor for changing periodicity of structures."""

    structure = traitlets.Instance(ase.Atoms, allow_none=True)

    def __init__(self, title="", **kwargs):
        self.title = title

        self.periodicity = ipw.RadioButtons(
            options=[
                ("3D (bulk systems)", "xyz"),
                ("2D (surfaces, slabs, ...)", "xy"),
                ("1D (wires)", "x"),
                ("0D (molecules)", "molecule"),
            ],
            value="xyz",
            layout={"width": "initial"},
        )
        self.apply_periodicity = ipw.Button(
            description="Apply",
            button_style="primary",
            layout={"width": "100px"},
        )
        self.apply_periodicity.on_click(self._select_periodicity)

        super().__init__(
            children=[
                ipw.HTML("""
                    <p>Select the periodicity of your system.</p>
                    <p style="font-weight: bold; color: #1f77b4;">NOTE:</p>

                    <ul style="padding-left: 2em; list-style-type: disc;">
                        <li>For <b>2D</b> systems (e.g., surfaces, slabs), the non-periodic direction must be the third lattice vector (z-axis).</li>
                        <li>For <b>1D</b> systems (e.g., wires), the periodic direction must be the first lattice vector (x-axis).</li>

                    </ul>
                """),
                self.periodicity,
                self.apply_periodicity,
            ],
            **kwargs,
        )

    def _select_periodicity(self, _=None):
        """Select periodicity."""
        periodicity_options = {
            "xyz": (True, True, True),
            "xy": (True, True, False),
            "x": (True, False, False),
            "molecule": (False, False, False),
        }
        new_structure = deepcopy(self.structure)
        new_structure.set_pbc(periodicity_options[self.periodicity.value])
        self.structure = None
        self.structure = deepcopy(new_structure)


class QEAppComputationalResourcesWidget(ipw.VBox):
    value = traitlets.Unicode(allow_none=True)
    nodes = traitlets.Int(default_value=1)
    cpus = traitlets.Int(default_value=1)

    def __init__(self, **kwargs):
        """Widget to setup the compute resources, which include the code,
        the number of nodes and the number of cpus.
        """
        self.code_selection = ComputationalResourcesWidget(
            include_setup_widget=False,
            fetch_codes=True,  # TODO resolve testing issues when set to `False`
            **kwargs,
        )
        self.code_selection.layout.width = "80%"

        self.num_nodes = ipw.BoundedIntText(
            value=1, step=1, min=1, max=1000, description="Nodes", width="10%"
        )
        self.num_cpus = ipw.BoundedIntText(
            value=1, step=1, min=1, description="CPUs", width="10%"
        )
        self.btn_setup_resource_detail = ipw.ToggleButton(description="More")
        self.btn_setup_resource_detail.observe(self._setup_resource_detail, "value")
        self._setup_resource_detail_output = ipw.Output(layout={"width": "500px"})

        # combine code, nodes and cpus
        children = [
            ipw.HBox(
                children=[
                    self.code_selection,
                    self.num_nodes,
                    self.num_cpus,
                    self.btn_setup_resource_detail,
                ]
            ),
            self._setup_resource_detail_output,
        ]
        super().__init__(children=children, **kwargs)

        self.resource_detail = ResourceDetailSettings()
        traitlets.dlink(
            (self.num_cpus, "value"), (self.resource_detail.ntasks_per_node, "value")
        )
        traitlets.link((self.code_selection, "value"), (self, "value"))

    def update_resources(self, change):
        if change["new"]:
            self.set_resource_defaults(load_code(change["new"]).computer)

    def set_resource_defaults(self, computer=None):
        if computer is None:
            self.num_nodes.disabled = True
            self.num_nodes.value = 1
            self.num_cpus.max = 1
            self.num_cpus.value = 1
            self.num_cpus.description = "CPUs"
        else:
            default_mpiprocs = computer.get_default_mpiprocs_per_machine()
            self.num_nodes.disabled = (
                True if computer.hostname == "localhost" else False
            )
            self.num_cpus.max = default_mpiprocs
            self.num_cpus.value = (
                1 if computer.hostname == "localhost" else default_mpiprocs
            )
            self.num_cpus.description = "CPUs"

    @property
    def parameters(self):
        return self.get_parameters()

    def get_parameters(self):
        """Return the parameters."""
        parameters = {
            "code": self.code_selection.value,
            "nodes": self.num_nodes.value,
            "cpus": self.num_cpus.value,
        }
        parameters.update(self.resource_detail.parameters)
        return parameters

    @parameters.setter
    def parameters(self, parameters):
        self.set_parameters(parameters)

    def set_parameters(self, parameters):
        """Set the parameters."""
        self.code_selection.value = parameters["code"]
        if "nodes" in parameters:
            self.num_nodes.value = parameters["nodes"]
        if "cpus" in parameters:
            self.num_cpus.value = parameters["cpus"]
        if "ntasks_per_node" in parameters:
            self.resource_detail.ntasks_per_node.value = parameters["ntasks_per_node"]
        if "cpus_per_task" in parameters:
            self.resource_detail.cpus_per_task.value = parameters["cpus_per_task"]
        if "max_wallclock_seconds" in parameters:
            self.resource_detail.max_wallclock_seconds.value = parameters[
                "max_wallclock_seconds"
            ]

    def _setup_resource_detail(self, _=None):
        with self._setup_resource_detail_output:
            clear_output()
            if self.btn_setup_resource_detail.value:
                self._setup_resource_detail_output.layout = {
                    "width": "500px",
                    "border": "1px solid gray",
                }

                children = [
                    self.resource_detail,
                ]
                display(*children)
            else:
                self._setup_resource_detail_output.layout = {
                    "width": "500px",
                    "border": "none",
                }


class ResourceDetailSettings(ipw.VBox):
    """Widget for setting the Resource detail."""

    prompt = ipw.HTML(
        """<div style="line-height:120%; padding-top:0px">
        <p style="padding-bottom:10px">
        Specify the parameters for the scheduler (only for advanced user). <br>
        These should be specified accordingly to the computer where the code will run.
        </p></div>"""
    )

    def __init__(self, **kwargs):
        self.ntasks_per_node = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            max=1000,
            description="ntasks-per-node",
            style={"description_width": "100px"},
        )
        self.cpus_per_task = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            description="cpus-per-task",
            style={"description_width": "100px"},
        )
        self.max_wallclock_seconds = ipw.BoundedIntText(
            value=3600 * 12,
            step=3600,
            min=60 * 10,
            max=3600 * 24,
            description="max seconds",
            style={"description_width": "100px"},
        )
        super().__init__(
            children=[
                self.prompt,
                self.ntasks_per_node,
                self.cpus_per_task,
                self.max_wallclock_seconds,
            ],
            **kwargs,
        )

    @property
    def parameters(self):
        return self.get_parameters()

    def get_parameters(self):
        """Return the parameters."""
        return {
            "ntasks_per_node": self.ntasks_per_node.value,
            "cpus_per_task": self.cpus_per_task.value,
            "max_wallclock_seconds": self.max_wallclock_seconds.value,
        }

    @parameters.setter
    def parameters(self, parameters):
        self.ntasks_per_node.value = parameters.get("ntasks_per_node", 1)
        self.cpus_per_task.value = parameters.get("cpus_per_task", 1)
        self.max_wallclock_seconds.value = parameters.get(
            "max_wallclock_seconds", 3600 * 12
        )

    def reset(self):
        """Reset the settings."""
        self.ntasks_per_node.value = 1
        self.cpus_per_task.value = 1
        self.max_wallclock_seconds.value = 3600 * 12


class ParallelizationSettings(ipw.VBox):
    """Widget for setting the parallelization settings."""

    prompt = ipw.HTML(
        """<div style="line-height:120%; padding-top:0px">
        <p style="padding-bottom:10px">
        Specify the number of k-points pools for the pw.x calculations (only for advanced user).
        </p></div>"""
    )

    def __init__(self, **kwargs):
        extra = {
            "style": {"description_width": "150px"},
            "layout": {"min_width": "180px"},
        }
        self.npool = ipw.BoundedIntText(
            value=1, step=1, min=1, max=128, description="Number of k-pools", **extra
        )
        self.override = ipw.Checkbox(
            escription="",
            indent=False,
            value=False,
            layout=ipw.Layout(max_width="20px"),
        )
        self.override.observe(self.set_visibility, "value")
        super().__init__(
            children=[
                ipw.HBox(
                    children=[self.override, self.prompt, self.npool],
                    layout=ipw.Layout(justify_content="flex-start"),
                ),
            ],
            **kwargs,
        )
        # set the default visibility of the widget
        self.npool.layout.display = "none"

    def set_visibility(self, change):
        if change["new"]:
            self.npool.layout.display = "block"
        else:
            self.npool.layout.display = "none"

    def reset(self):
        """Reset the parallelization settings."""
        self.npool.value = 1


class PwCodeResourceSetupWidget(QEAppComputationalResourcesWidget):
    """ComputationalResources Widget for the pw.x calculation."""

    nodes = traitlets.Int(default_value=1)

    def __init__(self, **kwargs):
        # By definition, npool must be a divisor of the total number of k-points
        # thus we can not set a default value here, or from the computer.
        self.parallelization = ParallelizationSettings()
        super().__init__(**kwargs)
        # add nodes and cpus into the children of the widget
        self.children += (self.parallelization,)

    def get_parallelization(self):
        """Return the parallelization settings."""
        parallelization = (
            {"npool": self.parallelization.npool.value}
            if self.parallelization.override.value
            else {}
        )
        return parallelization

    def set_parallelization(self, parallelization):
        """Set the parallelization settings."""
        if "npool" in parallelization:
            self.parallelization.override.value = True
            self.parallelization.npool.value = parallelization["npool"]

    def get_parameters(self):
        """Return the parameters."""
        parameters = super().get_parameters()
        parameters.update({"parallelization": self.get_parallelization()})
        return parameters

    def set_parameters(self, parameters):
        """Set the parameters."""
        super().set_parameters(parameters)
        if "parallelization" in parameters:
            self.set_parallelization(parameters["parallelization"])


class LoadingWidget(ipw.HBox):
    """Widget for displaying a loading spinner."""

    def __init__(self, message="Loading", **kwargs):
        self.message = ipw.Label(message)
        super().__init__(
            children=[
                ipw.Label(message),
                ipw.HTML(
                    value="<i class='fa fa-spinner fa-spin fa-2x fa-fw'/>",
                    layout=ipw.Layout(margin="12px 0 6px"),
                ),
            ],
            layout=ipw.Layout(
                justify_content="center",
                align_items="center",
                **kwargs.pop("layout", {}),
            ),
            **kwargs,
        )
        self.add_class("loading")


class LazyLoader(ipw.VBox):
    identifier = "widget"

    def __init__(self, widget_class, widget_kwargs=None, **kwargs):
        super().__init__(
            children=[LoadingWidget(f"Loading {self.identifier}")],
            **kwargs,
        )

        self._widget_class = widget_class
        self._widget_kwargs = widget_kwargs or {}

        self.rendered = False

    def set_widget_kwargs(self, kwargs):
        self._widget_kwargs = kwargs

    def render(self):
        if self.rendered:
            return
        self.widget = self._widget_class(**self._widget_kwargs)
        self.children = [self.widget]
        self.rendered = True


class LazyLoadedStructureImporter(ipw.VBox):
    """A wrapper that may be used to lazy-load a structure import widget."""

    warning_message = "This may take some time to load"

    structure = traitlets.Union(
        [
            traitlets.Instance(ase.Atoms),
            traitlets.Instance(orm_Data),
        ],
        allow_none=True,
    )

    def __init__(self, title=None, **kwargs):
        self.title = title or "Structure importer"

        render_button = ipw.Button(
            layout=ipw.Layout(margin="0 10px 0 0", width="auto"),
            description=f"Load {self.title} widget",
            icon="refresh",
        )
        render_button.on_click(self._on_render_click)

        super().__init__(
            children=[
                ipw.HBox(
                    layout=ipw.Layout(align_items="center"),
                    children=[
                        render_button,
                        ipw.HTML("<span class='warning'>WARNING: </span>"),
                        ipw.HTML(f"<span>{self.warning_message}</span>"),
                    ],
                ),
            ],
            **kwargs,
        )

        self.rendered = False

    def render(self):
        if self.rendered:
            return

        self.children = [LoadingWidget(f"Loading {self.title} widget")]

        widget = self._get_widget()
        ipw.dlink(
            (widget, "structure"),
            (self, "structure"),
        )

        self.children = [widget]

        self.rendered = True

    def _get_widget(self):
        raise NotImplementedError()

    def _on_render_click(self, _):
        self.render()


class LazyLoadedOptimade(LazyLoadedStructureImporter):
    warning_message = (
        "OPTIMADE may take some time to load depending on your internet connection"
    )

    def _get_widget(self):
        from aiidalab_widgets_base.databases import OptimadeQueryWidget

        return OptimadeQueryWidget(embedded=False)


class LazyLoadedStructureBrowser(LazyLoadedStructureImporter):
    warning_message = (
        "The browser may take some time to load depending on the size of your database"
    )

    def _get_widget(self):
        from aiida import orm
        from aiida_quantumespresso.data.hubbard_structure import HubbardStructureData
        from aiidalab_widgets_base.structures import StructureBrowserWidget

        return StructureBrowserWidget(
            title=self.title,
            query_types=(
                orm.StructureData,
                orm.CifData,
                HubbardStructureData,
            ),
        )


class CategorizedStructureExamplesWidget(StructureExamplesWidget):
    """Extended widget to provide categorized example structures."""

    def __init__(self, examples_by_category, title="", **kwargs):
        self.examples_by_category = examples_by_category
        self._category_buttons = ipw.ToggleButtons(
            options=list(examples_by_category.keys()),
            description="Category:",
        )
        self._category_buttons.observe(self._on_category_change, names="value")

        # Initialize with the first category
        initial_category = next(iter(examples_by_category.keys()))
        super().__init__(
            examples=examples_by_category[initial_category], title=title, **kwargs
        )

        self.children = [self._category_buttons, *list(self.children)]

    def _on_category_change(self, change):
        """Update the dropdown when the category changes."""
        new_category = change["new"]
        new_examples = self.examples_by_category.get(new_category, [])
        self._select_structure.options = self.get_example_structures(new_examples)


class LinkButton(ipw.HTML):
    disabled = traitlets.Bool(False)

    def __init__(
        self,
        description=None,
        link="",
        in_place=False,
        class_="",
        style_="",
        icon="",
        disabled=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        html = f"""
            <a
                role="button"
                href="{link}"
                target="{"_self" if in_place else "_blank"}"
                style="cursor: default; {style_}"
            >
        """
        if icon:
            html += f"<i class='fa fa-{icon}'></i>"

        html += f"{description}</a>"

        self.value = html

        self.add_class("jupyter-button")
        self.add_class("widget-button")
        self.add_class("link-button")
        self.add_class(class_)

        self.disabled = disabled

    @traitlets.observe("disabled")
    def _on_disabled(self, change):
        if change["new"]:
            self.add_class("disabled")
        else:
            self.remove_class("disabled")


class TableWidget(anywidget.AnyWidget):
    _esm = """
    function render({ model, el }) {
        let domElement = document.createElement("div");
        el.classList.add("custom-table");
        let selectedIndices = [];

        function drawTable() {
            const data = model.get("data");
            domElement.innerHTML = "";
            let innerHTML = '<table><tr>' + data[0].map(header => `<th>${header}</th>`).join('') + '</tr>';

            for (let i = 1; i < data.length; i++) {
                innerHTML += '<tr>' + data[i].map(cell => `<td>${cell}</td>`).join('') + '</tr>';
            }

            innerHTML += "</table>";
            domElement.innerHTML = innerHTML;

            const rows = domElement.querySelectorAll('tr');
            rows.forEach((row, index) => {
                if (index > 0) {
                    row.addEventListener('click', () => {
                        const rowIndex = index - 1;
                        if (selectedIndices.includes(rowIndex)) {
                            selectedIndices = selectedIndices.filter(i => i !== rowIndex);
                            row.classList.remove('selected-row');
                        } else {
                            selectedIndices.push(rowIndex);
                            row.classList.add('selected-row');
                        }
                        model.set('selected_rows', [...selectedIndices]);
                        model.save_changes();
                    });

                    row.addEventListener('mouseover', () => {
                        if (!row.classList.contains('selected-row')) {
                            row.classList.add('hover-row');
                        }
                    });

                    row.addEventListener('mouseout', () => {
                        row.classList.remove('hover-row');
                    });
                }
            });
        }

        function updateSelection() {
            const newSelection = model.get("selected_rows");
            selectedIndices = [...newSelection]; // Synchronize the JavaScript state with the Python state
            const rows = domElement.querySelectorAll('tr');
            rows.forEach((row, index) => {
                if (index > 0) {
                    if (selectedIndices.includes(index - 1)) {
                        row.classList.add('selected-row');
                    } else {
                        row.classList.remove('selected-row');
                    }
                }
            });
        }

        drawTable();
        model.on("change:data", drawTable);
        model.on("change:selected_rows", updateSelection);
        el.appendChild(domElement);
    }
    export default { render };
    """
    _css = """
    .custom-table table, .custom-table th, .custom-table td {
        border: 1px solid black;
        border-collapse: collapse;
        text-align: left;
        padding: 4px;
    }
    .custom-table th, .custom-table td {
        min-width: 50px;
        word-wrap: break-word;
    }
    .custom-table table {
        width: 70%;
        font-size: 1.0em;
    }
    /* Hover effect with light gray background */
    .custom-table tr.hover-row:not(.selected-row) {
        background-color: #f0f0f0;
    }
    /* Selected row effect with light green background */
    .custom-table tr.selected-row {
        background-color: #DFF0D8;
    }
    """
    data = traitlets.List().tag(sync=True)
    selected_rows = traitlets.List().tag(sync=True)


class HBoxWithUnits(ipw.HBox):
    def __init__(self, widget: ipw.ValueWidget, units: str, **kwargs):
        super().__init__(
            children=[
                widget,
                ipw.HTML(units),
            ],
            layout=ipw.Layout(
                align_items="center",
                grid_gap="2px",
            ),
            **kwargs,
        )


class ArchiveImporter(ipw.VBox):
    GITHUB = "https://github.com"
    INFO_TEMPLATE = "{} <i class='fa fa-spinner fa-spin'></i>"

    def __init__(
        self,
        repo: str,
        tag: str,
        archive_list: str,
        archives_dir: str,
        logger: t.Optional[dict] = None,
        **kwargs,
    ):
        refs = "refs/tags"
        raw = f"https://raw.githubusercontent.com/{repo}/{refs}/{tag}"
        self.archive_list_url = f"{raw}/{archive_list}"
        self.archives_url = f"{self.GITHUB}/{repo}/raw/{refs}/{tag}/{archives_dir}"
        if logger:
            self.logger_placeholder = logger.get("placeholder", "")
            self.clear_log_on_import = logger.get("clear_on_import", False)
            self.logger = RollingOutput()
            self.logger.value = self.logger_placeholder
        super().__init__(children=[LoadingWidget()], **kwargs)

    def render(self):
        self.selector = ipw.SelectMultiple(
            options=[],
            description="Examples:",
            rows=10,
            style={"description_width": "initial"},
            layout=ipw.Layout(width="auto"),
        )

        self.import_button = ipw.Button(
            description="Import",
            button_style="success",
            layout=ipw.Layout(width="fit-content"),
            icon="download",
        )
        ipw.dlink(
            (self.selector, "value"),
            (self.import_button, "disabled"),
            lambda value: not value,
        )
        self.import_button.on_click(self.import_archives)

        self.info = ipw.HTML()

        history_link = LinkButton(
            description="Calculation history",
            link="./calculation_history.ipynb",
            icon="list",
            class_="mod-primary",
            style_="color: white;",
            layout=ipw.Layout(
                width="fit-content",
                margin="2px 0 2px auto",
            ),
        )

        accordion = None
        if self.logger:
            accordion = ipw.Accordion(children=[self.logger])
            accordion.set_title(0, "Archive import log")
            accordion.selected_index = None

        self.children = [
            self.selector,
            ipw.HBox(
                [
                    self.import_button,
                    self.info,
                    history_link,
                ],
                layout=ipw.Layout(
                    margin="2px 2px 4px 68px",
                    align_items="center",
                    grid_gap="4px",
                ),
            ),
            accordion or ipw.Box,
        ]

        self.selector.options = self.get_options()

    def get_options(self) -> list[tuple[str, str]]:
        response: req.Response = req.get(self.archive_list_url)
        if not response.ok:
            self.info.value = "Failed to fetch archive list"
            return []
        if archives := response.content.decode("utf-8").strip().split("\n"):
            return [(archive, archive.split("-")[0].strip()) for archive in archives]
        self.info.value = "No archives found"
        return []

    def import_archives(self, _):
        self.import_button.disabled = True
        if self.logger and self.clear_log_on_import:
            self.logger.value = ""
        for filename in self.selector.value:
            self.import_archive(filename)
        self.import_button.disabled = False

    def import_archive(self, filename: str):
        self.info.value = self.INFO_TEMPLATE.format(f"Importing {filename}")
        file_url = f"{self.archives_url}/{filename}"
        process = subprocess.Popen(
            ["verdi", "archive", "import", "-v", "critical", file_url],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self._report(filename, stdout, stderr)

    def _report(self, filename: str, stdout: str, stderr: str):
        if stderr and "Success" not in stdout:
            self.info.value = f"Failed to import {filename}"
        if self.logger:
            self.logger.value += stdout
            if stderr:
                if "Success" not in stdout:
                    self.logger.value += "\n[ERROR]"
                    self.info.value = "Error -> see log for details"
                else:
                    self.info.value = ""
                self.logger.value += f"\n{stderr}"


class ShakeNBreakEditor(ipw.VBox):
    structure = traitlets.Instance(ase.Atoms, allow_none=True)
    selection = traitlets.List(traitlets.Int)
    structure_node = traitlets.Instance(orm_Data, allow_none=True, read_only=True)

    def __init__(self, title="Editor ShakeNbreak"):
        self.title = title
        self._status_message = StatusHTML()
        self.num_nearest_neighbours = ipw.IntText(
            description="Num. neighbours atoms:",
            value=8,
            step=1,
            style={"description_width": "150px"},
        )
        self.defect_position = ipw.Text(
            description="Defect atom:",
            style={"description_width": "150px"},
        )
        # Only select one atom!
        self.btn_defect_position = ipw.Button(
            description="From selection",
            style={"description_width": "150px"},
        )
        # Center of the selection
        self.btn_defect_position_vac = ipw.Button(
            description="From selection",
            style={"description_width": "150px"},
        )
        self.vacancy_coords = ipw.Text(
            description="Vacancy coords:",
            style={"description_width": "150px"},
        )
        self.btn_defect_position.on_click(self._defect_position)
        self.btn_defect_position_vac.on_click(self._defect_position_vac)
        self.distortion_factor = ipw.BoundedFloatText(
            value=1.0,
            min=0.2,
            max=1.8,
            step=0.1,
            description="Distortion factor:",
            style={"description_width": "150px"},
        )
        self.btn_apply_bond_distortion = ipw.Button(
            description="Apply bond distortion",
            button_style="primary",
            disabled=False,
        )
        self.btn_apply_bond_distortion.on_click(self._apply_bond_distortion)
        self.selected_atoms = ipw.Text(
            description="Select atoms:", value="", style={"description_width": "150px"}
        )
        self.btn_selected_atoms = ipw.Button(
            description="From selection", style={"description_width": "150px"}
        )
        self.btn_selected_atoms.on_click(self._selected_atoms)
        self.wrong_syntax = ipw.HTML(
            value="""<i class="fa fa-times" style="color:red;font-size:2em;" ></i> wrong syntax""",
            layout={"visibility": "hidden"},
        )
        self.radial_cutoff = ipw.FloatText(
            description="Radial cutoff distance (Å):",
            value=3.0,
            style={"description_width": "150px"},
        )
        self.btn_apply_random_distorion_all = ipw.Button(
            description="Apply to all atoms",
            button_style="primary",
            style={"description_width": "250px"},
        )
        self.btn_apply_random_distorion_all.on_click(self._apply_random_distortion_all)
        self.btn_apply_random_distortion = ipw.Button(
            description="Apply locally",
            button_style="primary",
            style={"description_width": "250px"},
        )
        self.btn_apply_random_distortion.on_click(self._apply_random_distortion)
        super().__init__(
            children=[
                ipw.HTML(
                    """
                    <p>
                    Apply bond distortions or random displacements to explore metastable structures of point defects in solids.
                    This editor allows you to manipulate atomic positions and study defect behaviors in materials.
                    </p>
                    """
                ),
                ipw.HTML(
                    "<h4>Define defect position:</h4>",
                ),
                ipw.HTML(
                    """
                    <p>
                    To apply bond distortions, you can define either:<br>
                    - A <strong>defect atom</strong>: Select an atom in the viewer to mark it as a defect.<br>
                    - A <strong>vacancy position</strong>: Select the center of mass of the selection to define the vacancy position.
                    </p>
                    """
                ),
                ipw.HBox(
                    [
                        self.defect_position,
                        self.btn_defect_position,
                    ]
                ),
                ipw.HBox(
                    [
                        self.vacancy_coords,
                        self.btn_defect_position_vac,
                    ]
                ),
                ipw.HTML(
                    "<h4>Bond distortion around defect:</h4>",
                ),
                ipw.HTML(
                    """
                    <p>
                    Specify the number of neighboring atoms to apply the bond distortion.
                    Set a distortion factor to adjust atomic positions:<br>
                    - A value less than 1.0 shrinks the positions (e.g., 0.8 for a 20% reduction). <br>
                    - A value greater than 1.0 expands the positions (e.g., 1.8 for an 80% increase).
                    </p>
                    """
                ),
                ipw.VBox(
                    [
                        self.num_nearest_neighbours,
                        self.distortion_factor,
                        self.btn_apply_bond_distortion,
                    ]
                ),
                ipw.HTML(
                    "<h4>Random displacements to all atoms or selected ones:</h4>",
                ),
                ipw.HTML(
                    """
                    <p>
                    Define atoms for local random displacement around the defect or vacancy, and set the radial cutoff distance (in Å) to determine neighboring atoms for distance checks.
                    </p>
                    """
                ),
                ipw.HBox(
                    [
                        self.selected_atoms,
                        self.btn_selected_atoms,
                        self.wrong_syntax,
                    ]
                ),
                self.radial_cutoff,
                ipw.HBox(
                    [
                        self.btn_apply_random_distortion,
                        self.btn_apply_random_distorion_all,
                    ],
                ),
                self._status_message,
                ipw.HTML(
                    """
                    <p>
                    Uses the <a href="https://github.com/SMTG-Bham/ShakeNBreak" target="_blank">ShakeNBreak</a> package.
                    </p>
                    """
                ),
            ]
        )

    def sel2com(self):
        """Get center of mass of the selection."""
        if self.selection:
            com = np.average(
                self.structure[self.selection].get_scaled_positions(), axis=0
            )
        else:
            com = [0, 0, 0]

        return com

    def str2vec(self, string):
        return np.array(list(map(float, string.split())))

    def vec2str(self, vector):
        return (
            str(round(vector[0], 2))
            + " "
            + str(round(vector[1], 2))
            + " "
            + str(round(vector[2], 2))
        )

    def _defect_position_vac(self, _=None):
        """Define vaccancy coordinates."""
        self.vacancy_coords.value = self.vec2str(self.sel2com())

    def _defect_position(self, _=None):
        """Define vaccancy coordinates."""
        if len(self.selection) != 1:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select one atom first.</strong>
            </div>
            """
        else:
            self.defect_position.value = list_to_string_range(self.selection)

    def _selected_atoms(self, _=None):
        """Selected atoms to displace."""
        self.selected_atoms.value = list_to_string_range(self.selection)

    def _apply_bond_distortion(self, _=None):
        if not self.defect_position.value and not self.vacancy_coords.value:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select the position of the defect or vacancy.</strong>
            </div>
            """
        else:
            if self.defect_position.value == "":
                site_index = None
            else:
                site_index = int(self.defect_position.value)
            if self.vacancy_coords.value == "":
                frac_coords = None
            else:
                frac_coords = self.str2vec(self.vacancy_coords.value)

            self._apply_distortion(
                distort,
                num_nearest_neighbours=self.num_nearest_neighbours.value,
                site_index=site_index,
                distortion_factor=self.distortion_factor.value,
                frac_coords=frac_coords,
            )

    def _apply_random_distortion(self, _=None):
        if not self.defect_position.value and not self.vacancy_coords.value:
            self._status_message.message = """
            <div class="alert alert-info">
            <strong>Please select the position of the defect or vacancy.</strong>
            </div>
            """
        else:
            if self.defect_position.value == "":
                site_index = None
            else:
                site_index = int(self.defect_position.value)
            if self.vacancy_coords.value == "":
                frac_coords = None
            else:
                frac_coords = self.str2vec(self.vacancy_coords.value)

            active_atoms, syntax_ok = string_range_to_list(self.selected_atoms.value)
            if not active_atoms:
                active_atoms = None
            if not syntax_ok:
                self.wrong_syntax.layout.visibility = "visible"
            else:
                self.wrong_syntax.layout.visibility = "hidden"

                self._apply_distortion(
                    local_mc_rattle,
                    site_index=site_index,
                    frac_coords=frac_coords,
                    active_atoms=active_atoms,
                    nbr_cutoff=self.radial_cutoff.value,
                )

    def _apply_random_distortion_all(self, _=None):
        self._apply_distortion(
            rattle,
            active_atoms=None,
            nbr_cutoff=self.radial_cutoff.value,
        )

    def _apply_distortion(self, distortion_func, **kwargs):
        pymatgen_ase = AseAtomsAdaptor()
        pymatgen_structure = pymatgen_ase.get_structure(self.structure)
        struc_distorted = distortion_func(structure=pymatgen_structure, **kwargs)
        periodicity = self.structure.pbc
        if "distorted_structure" in struc_distorted:
            atoms = pymatgen_ase.get_atoms(struc_distorted["distorted_structure"])
        else:
            atoms = pymatgen_ase.get_atoms(struc_distorted)
        atoms.set_pbc(periodicity)
        self.structure = atoms



import traitlets as tl


class MetaHasTraitsLast(tl.MetaHasTraits):
    """A metaclass that ensures that `HasTraits` is pushed to the end of the MRO.

    This is useful when injecting trait-ful mixins into the MRO of a class."""

    def __new__(cls, name, bases, classdict):
        bases = (
            *filter(lambda base: base is not tl.HasTraits, bases),
            tl.HasTraits,
        )
        return super().__new__(cls, name, bases, classdict)


class Model(tl.HasTraits, metaclass=MetaHasTraitsLast):
    """A parent class for all MVC models.

    The class extends `traitlet`'s `HasTraits` and uses a metaclass to
    ensure that the MRO is reordered such that `HasTraits`, which may
    be extended by other bases (mixins), is pushed to the end of the
    MRO to avoid conflicts - see C3 linearization.
    """



import time

import pytest
from click.testing import CliRunner, Result

import aiida
import aiidalab_qe.__main__ as cli

# To learn more about testing click applications, see: https://click.palletsprojects.com/en/8.1.x/testing/


@pytest.mark.slow
def test_download_and_install_pseudos(tmp_path, aiida_profile, monkeypatch):
    """Test download pseudos to a specified directory and install them.
    And test install pseudos without download.
    Compare the time elapsed between the two methods.

    Act: Run the command `aiidalab-qe download-pseudos --dest <tmp_path>`.
    Assert: All pseudos are downloaded to the specified directory.

    Act: Run the command `aiidalab-qe install-pseudos --source <tmp_path>`.
    Assert: All pseudos are installed from the specified directory.

    Act: Run the command `aiidalab-qe install-pseudos`.
    Assert: All pseudos are installed from the default directory, but slow.

    Note: this test is slow, it takes about ~30 seconds to run.
    """
    from aiidalab_qe.setup.pseudos import (
        PSEUDODOJO_VERSION,
        SSSP_VERSION,
        pseudos_to_install,
    )

    # Mock the EXPECTED_PSEUDOS to speed up the test
    MOCK_EXPECTED_PSEUDOS = {
        f"SSSP/{SSSP_VERSION}/PBE/efficiency",
        f"PseudoDojo/{PSEUDODOJO_VERSION}/PBEsol/SR/standard/upf",
        f"PseudoDojo/{PSEUDODOJO_VERSION}/PBE/SR/stringent/upf",
        f"PseudoDojo/{PSEUDODOJO_VERSION}/PBEsol/SR/stringent/upf",
    }
    # mock the EXPECTED_PSEUDOS
    monkeypatch.setattr(
        "aiidalab_qe.setup.pseudos.EXPECTED_PSEUDOS",
        MOCK_EXPECTED_PSEUDOS,
    )

    # clean the profile database
    aiida_profile.clear_profile()

    runner: CliRunner = CliRunner()

    start = time.time()
    result: Result = runner.invoke(cli.download_pseudos, ["--dest", str(tmp_path)])
    download_time = time.time() - start

    assert result.exit_code == 0
    assert "Pseudopotentials are downloaded!" in result.output

    files = [f for f in tmp_path.glob("**/*") if f.is_file()]

    for pseudo in MOCK_EXPECTED_PSEUDOS:
        filename = f"{pseudo.replace('/', '_')}.aiida_pseudo"
        assert filename in [f.name for f in files]

    # Install the pseudos from the tmp_path

    # Check that the pseudos are not installed yet
    assert len(pseudos_to_install()) == len(MOCK_EXPECTED_PSEUDOS)

    profile = aiida.get_profile()

    start = time.time()
    result: Result = runner.invoke(
        cli.install_pseudos, ["--source", tmp_path, "--profile", profile.name]
    )
    install_time = time.time() - start

    print(f"Download time: {download_time}, install time: {install_time}")

    assert result.exit_code == 0
    assert "Pseudopotentials are installed!" in result.output

    assert len(pseudos_to_install()) == 0

    # clean the profile database for the next test
    aiida_profile.clear_profile()

    # Check that the pseudos are not installed yet
    assert len(pseudos_to_install()) == len(MOCK_EXPECTED_PSEUDOS)

    profile = aiida.get_profile()
    start = time.time()
    result: Result = runner.invoke(cli.install_pseudos, ["--profile", profile.name])
    install_without_download_time = time.time() - start
    print(f"Install without download time: {install_without_download_time}")

    assert result.exit_code == 0
    assert "Pseudopotentials are installed!" in result.output

    assert len(pseudos_to_install()) == 0

    # Check the install without download is slower than install with download
    assert install_without_download_time > install_time



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from aiidalab_qe.common.bands_pdos.utils import (
    find_max_in_range,
    find_max_up_and_down,
    prepare_combined_plotly_traces,
)


class BandsPdosPlotly:
    SETTINGS = {
        "axis_linecolor": "#111111",
        "bands_linecolor": "#111111",
        "bands_up_linecolor": "rgba(205, 0, 0, 0.4)",  # Red Opacitiy 40%
        "bands_down_linecolor": "rgba(72, 118, 255, 0.4)",  # Blue Opacitiy 40%
        "combined_plot_height": 600,
        "combined_plot_width": 900,
        "combined_column_widths": [0.7, 0.3],
        "bands_plot_height": 600,
        "bands_plot_width": 850,
        "pdos_plot_height": 600,
        "pdos_plot_width": 850,
        "vertical_linecolor": "#111111",
        "horizontal_linecolor": "#111111",
        "vertical_range_bands": [-10, 10],
        "horizontal_range_pdos": [-10, 10],
    }

    def __init__(
        self,
        bands_data=None,
        external_bands_data=None,
        pdos_data=None,
        bands_projections_data=None,
        plot_settings=None,
    ):
        self.bands_data = bands_data
        self.external_bands_data = external_bands_data
        self.pdos_data = pdos_data
        self.project_bands = bands_projections_data
        self.plot_settings = plot_settings or {}

        self.fermi_energy = self._get_fermi_energy()

        # Plotly axis objects
        self._bands_xaxis = self._band_xaxis()
        self._bands_yaxis = self._band_yaxis()
        self._pdos_xaxis = self._pdos_xaxis()
        self._pdos_yaxis = self._pdos_yaxis()

        # Plotly figure object
        self.figure_object = self._get_bandspdos_plot()

    @property
    def plot_type(self):
        """Define the plot type."""
        if self.bands_data and self.pdos_data:
            return "combined"
        elif self.bands_data:
            return "bands"
        elif self.pdos_data:
            return "pdos"

    @property
    def bandspdosfigure(self):
        return self.figure_object

    def _get_fermi_energy(self):
        """Function to return the Fermi energy information depending on the data available."""
        fermi_data = {}
        if self.pdos_data:
            if "fermi_energy_up" in self.pdos_data:
                fermi_data["fermi_energy_up"] = self.pdos_data["fermi_energy_up"]
                fermi_data["fermi_energy_down"] = self.pdos_data["fermi_energy_down"]
            else:
                fermi_data["fermi_energy"] = self.pdos_data["fermi_energy"]
        else:
            if "fermi_energy_up" in self.bands_data:
                fermi_data["fermi_energy_up"] = self.bands_data["fermi_energy_up"]
                fermi_data["fermi_energy_down"] = self.bands_data["fermi_energy_down"]
            else:
                fermi_data["fermi_energy"] = self.bands_data["fermi_energy"]
        return fermi_data

    def _band_xaxis(self):
        """Function to return the xaxis for the bands plot."""

        if not self.bands_data:
            return None
        paths = self.bands_data.get("paths")
        bandxaxis = go.layout.XAxis(
            title="k-points",
            range=[0, paths[-1]["x"][-1]],
            showgrid=True,
            showline=True,
            tickmode="array",
            fixedrange=False,
            tickvals=self.bands_data["pathlabels"][1],  # ,self.band_labels[1],
            ticktext=self.bands_data["pathlabels"][0],  # self.band_labels[0],
            showticklabels=True,
            linecolor=self.SETTINGS["axis_linecolor"],
            mirror=True,
            linewidth=2,
            type="linear",
        )

        return bandxaxis

    def _band_yaxis(self):
        """Function to return the yaxis for the bands plot."""

        if not self.bands_data:
            return None

        bandyaxis = go.layout.YAxis(
            title={"text": "Electronic Bands (eV)", "standoff": 1},
            side="left",
            showgrid=True,
            showline=True,
            zeroline=True,
            range=self.SETTINGS["vertical_range_bands"],
            fixedrange=False,
            automargin=True,
            ticks="inside",
            linewidth=2,
            linecolor=self.SETTINGS["axis_linecolor"],
            tickwidth=2,
            zerolinewidth=2,
        )

        return bandyaxis

    def _pdos_xaxis(self):
        """Function to return the xaxis for the pdos plot."""

        if not self.pdos_data:
            return None
        # For combined plot
        axis_settings = {
            "showgrid": True,
            "showline": True,
            "mirror": "ticks",
            "ticks": "inside",
            "linewidth": 2,
            "tickwidth": 2,
            "linecolor": self.SETTINGS["axis_linecolor"],
            "title": "Density of states",
            "side": "bottom",
            "automargin": True,
        }

        if self.plot_type != "combined":
            axis_settings["title"] = "Density of states (eV)"
            axis_settings["range"] = self.SETTINGS["horizontal_range_pdos"]
            axis_settings.pop("side")
            axis_settings.pop("automargin")

        return go.layout.XAxis(**axis_settings)

    def _pdos_yaxis(self):
        """Function to return the yaxis for the pdos plot."""

        if not self.pdos_data:
            return None

        axis_settings = {
            "showgrid": True,
            "showline": True,
            "side": "right" if self.plot_type == "combined" else "left",
            "mirror": "ticks",
            "ticks": "inside",
            "linewidth": 2,
            "tickwidth": 2,
            "linecolor": self.SETTINGS["axis_linecolor"],
            "zerolinewidth": 2,
        }

        return go.layout.YAxis(**axis_settings)

    def _get_bandspdos_plot(self):
        """Function to return the bands plot widget."""

        fig = self._create_fig()

        self.adding_bands_traces(fig)
        self.adding_pdos_traces(fig)
        self.adding_projected_bands(fig)

        if self.plot_type == "combined":
            self._customize_combined_layout(fig)
        else:
            self._customize_single_layout(fig)

        return go.FigureWidget(fig)

    def adding_bands_traces(self, fig):
        if self.bands_data:
            bands_trace_settings = self.plot_settings.get("bands_trace_settings", None)
            self._add_band_traces(
                fig, bands_data=self.bands_data, trace_settings=bands_trace_settings
            )

            band_labels = self.bands_data.get("pathlabels")
            for label in band_labels[1]:
                fig.add_vline(
                    x=label,
                    line={"color": self.SETTINGS["vertical_linecolor"], "width": 1},
                )
            if self.external_bands_data:
                for key, bands_data in self.external_bands_data.items():
                    trace_settings = bands_data.get("trace_settings", {})
                    trace_settings.setdefault("name", key)
                    self._add_band_traces(
                        fig,
                        bands_data=bands_data,
                        trace_settings=trace_settings,
                    )

    def adding_pdos_traces(self, fig):
        if self.pdos_data:
            self._add_pdos_traces(fig)
            if self.plot_type == "pdos":
                fig.add_vline(
                    x=0,
                    line={
                        "color": self.SETTINGS["vertical_linecolor"],
                        "width": 1,
                        "dash": "dot",
                    },
                )

    def adding_projected_bands(self, fig):
        if self.project_bands:
            self._add_projection_traces(fig)

    def _create_fig(self):
        """Create a plotly figure.

        The figure layout is different depending on the plot type.
        """
        if self.plot_type != "combined":
            return go.Figure()

        fig = make_subplots(
            rows=1,
            cols=2,
            shared_yaxes=True,
            column_widths=self.SETTINGS["combined_column_widths"],
            horizontal_spacing=0.015,
        )
        return fig

    def _add_traces_to_fig(self, fig, traces, col):
        """Add a list of traces to a figure."""
        if self.plot_type == "combined":
            rows = [1] * len(traces)
            cols = [col] * len(traces)
            fig.add_traces(traces, rows=rows, cols=cols)
        else:
            fig.add_traces(traces)

    def _add_band_traces(self, fig, bands_data, trace_settings=None):
        """Generate the band traces and add them to the figure."""
        from copy import deepcopy

        trace_settings = trace_settings or {}
        name = trace_settings.pop("name", "Bands")
        trace_settings.setdefault("dash", "solid")
        trace_settings.setdefault("shape", "linear")
        colors = {
            (True, 0): self.SETTINGS["bands_up_linecolor"],
            (True, 1): self.SETTINGS["bands_down_linecolor"],
            (False, 0): self.SETTINGS["bands_linecolor"],
        }
        fermi_energy_mapping = {
            (False, 0): self.fermi_energy.get("fermi_energy_up", None),
            (False, 1): self.fermi_energy.get("fermi_energy_down", None),
        }

        trace_name_mapping = {
            (False, 0): f"{name}",  # Base case: non-spin-polarized
            (True, 0): f"{name} (↑)",  # Spin-up case
            (True, 1): f"{name} (↓)",  # Spin-down case
        }

        # Convert paths to a list of Scatter objects
        scatter_objects = []

        spin_polarized = 1 in bands_data["band_type_idx"]
        for spin in [0, 1]:
            trace_settings_spin = deepcopy(trace_settings)
            # In case of non-spin-polarized or SOC calculations, the spin index is only 0
            if spin not in bands_data["band_type_idx"]:
                continue

            x_bands = np.array(bands_data["x"]).reshape(1, -1)
            # New shape: (number of bands, number of kpoints)
            y_bands = bands_data["y"][:, bands_data["band_type_idx"] == spin].T
            # Concatenate the bands and prepare the traces
            x_bands_comb, y_bands_comb = prepare_combined_plotly_traces(
                x_bands, y_bands
            )

            fermi_energy = fermi_energy_mapping.get(
                ("fermi_energy" in self.fermi_energy, spin),
                self.fermi_energy.get("fermi_energy"),
            )
            trace_settings_spin.setdefault("color", colors[(spin_polarized, spin)])
            scatter_objects.append(
                go.Scattergl(
                    x=x_bands_comb,
                    y=y_bands_comb - fermi_energy,
                    mode="lines",
                    line=trace_settings_spin,
                    showlegend=spin_polarized,
                    name=trace_name_mapping[(spin_polarized, spin)],
                )
            )

        self._add_traces_to_fig(fig, scatter_objects, 1)

    def _add_pdos_traces(self, fig):
        # Extract DOS data
        dos_data = self.pdos_data["dos"]

        # Pre-allocate memory for Scatter objects
        num_traces = len(dos_data)
        scatter_objects = [None] * num_traces

        # dictionary with keys (bool(spin polarized), bool(spin up))
        fermi_energy_spin_mapping = {
            (False, True): self.fermi_energy.get("fermi_energy_up", None),
            (False, False): self.fermi_energy.get("fermi_energy_down", None),
        }

        # Vectorize Scatter object creation
        for i, trace in enumerate(dos_data):
            dos_np = np.array(trace["x"])
            fill = "tozerox" if self.plot_type == "combined" else "tozeroy"
            fermi_energy = fermi_energy_spin_mapping.get(
                ("fermi_energy" in self.fermi_energy, trace["label"].endswith("(↑)")),
                self.fermi_energy.get("fermi_energy"),
            )

            x_data = (
                trace["y"] if self.plot_type == "combined" else dos_np - fermi_energy
            )
            y_data = (
                dos_np - fermi_energy if self.plot_type == "combined" else trace["y"]
            )
            scatter_objects[i] = go.Scattergl(  # type: ignore
                x=x_data,
                y=y_data,
                fill=fill,
                name=trace["label"],
                line={
                    "color": trace["borderColor"],
                    "shape": "linear",
                },
                legendgroup=trace["label"],
            )

        self._add_traces_to_fig(fig, scatter_objects, 2)

    def _prepare_bands_projection_traces_data(self, projected_bands):
        """
        Prepares data for adding or updating projected band traces.

        Args:
            projected_bands (list[dict]): List of projected bands data.

        Returns:
            list[dict]: List of dictionaries containing x, y (with Fermi offset), color, and label.
        """
        spin_mapping = {
            (False, True): self.fermi_energy.get("fermi_energy_up", None),
            (False, False): self.fermi_energy.get("fermi_energy_down", None),
        }

        prepared_data = []
        for proj_bands in projected_bands:
            fermi_energy = spin_mapping.get(
                (
                    "fermi_energy" in self.fermi_energy,
                    proj_bands["label"].endswith("(↑)"),
                ),
                self.fermi_energy.get("fermi_energy"),
            )
            prepared_data.append(
                {
                    "x": proj_bands["x"],
                    "y": np.array(proj_bands["y"]) - fermi_energy,
                    "color": proj_bands["color"],
                    "label": proj_bands["label"],
                }
            )
        return prepared_data

    def _add_projection_traces(self, fig):
        """Function to add the projected bands traces to the bands plot."""
        prepared_data = self._prepare_bands_projection_traces_data(self.project_bands)

        scatter_objects = [
            go.Scattergl(
                x=data["x"],
                y=data["y"],
                fill="toself",
                legendgroup=data["label"],
                mode="lines",
                line={"width": 0, "color": data["color"]},
                name=data["label"],
                # If PDOS is present, use those legend entries
                showlegend=True if self.plot_type == "bands" else False,
            )
            for data in prepared_data
        ]

        self._add_traces_to_fig(fig, scatter_objects, 1)

    def update_projected_bands_thickness(self, fig):
        """Update the projected bands thickness."""
        prepared_data = self._prepare_bands_projection_traces_data(self.project_bands)

        # Create a mapping from labels to y-data for efficient updates
        y_data_by_label = {data["label"]: data["y"] for data in prepared_data}

        with fig.batch_update():
            for trace in fig.data:
                if trace.xaxis == "x" and trace.legendgroup in y_data_by_label:
                    trace.y = y_data_by_label[trace.legendgroup]

    def _customize_combined_layout(self, fig):
        self._customize_layout(fig, self._bands_xaxis, self._bands_yaxis)
        self._customize_layout(fig, self._pdos_xaxis, self._pdos_yaxis, col=2)
        fig.update_layout(
            legend={"xanchor": "left", "x": 1.06},
            height=self.SETTINGS["combined_plot_height"],
            width=self.SETTINGS["combined_plot_width"],
            plot_bgcolor="white",
        )
        self._update_dos_layout(fig)

    def _customize_layout(self, fig, xaxis, yaxis, row=1, col=1):
        fig.update_xaxes(patch=xaxis, row=row, col=col)
        fig.update_yaxes(patch=yaxis, row=row, col=col, showticklabels=True)
        fig.add_hline(
            y=0,
            line={
                "color": self.SETTINGS["horizontal_linecolor"],
                "width": 1,
                "dash": "dot",
            },
            row=row,
            col=col,
        )

    def _customize_single_layout(self, fig):
        xaxis = getattr(self, f"_{self.plot_type}_xaxis")
        yaxis = getattr(self, f"_{self.plot_type}_yaxis")

        fig.update_layout(
            xaxis=xaxis,
            yaxis=yaxis,
            plot_bgcolor="white",
            height=self.SETTINGS[f"{self.plot_type}_plot_height"],
            width=self.SETTINGS[f"{self.plot_type}_plot_width"],
        )
        self._update_dos_layout(fig)

    def _update_dos_layout(self, fig):
        def update_layout_spin_polarized(
            x_data_up,
            y_data_up,
            x_data_down,
            y_data_down,
            x_min,
            x_max,
            update_func,
            layout_type,
        ):
            most_negative_down, max_up = find_max_up_and_down(
                x_data_up, y_data_up, x_data_down, y_data_down, x_min, x_max
            )
            if layout_type == "layout":
                update_func(yaxis={"range": [most_negative_down * 1.10, max_up * 1.10]})
            elif layout_type == "xaxes":
                update_func(
                    patch={"range": [most_negative_down * 1.10, max_up * 1.10]},
                    row=1,
                    col=2,
                )

        def update_layout_non_spin_polarized(
            total_dos_xdata, total_dos_ydata, x_min, x_max, update_func, layout_type
        ):
            max_value = find_max_in_range(
                total_dos_xdata, total_dos_ydata, x_min, x_max
            )
            if layout_type == "layout":
                update_func(yaxis={"range": [0, max_value * 1.10]})
            elif layout_type == "xaxes":
                update_func(patch={"range": [0, max_value * 1.10]}, row=1, col=2)

        def get_x_min_max(fermi_energy):
            return (
                self.SETTINGS["horizontal_range_pdos"][0] + fermi_energy,
                self.SETTINGS["horizontal_range_pdos"][1] + fermi_energy,
            )

        def handle_spin_polarization(fermi_energy, update_func, layout_type):
            spin_polarized = "(↑)" in self.pdos_data["dos"][0]["label"]
            if not spin_polarized:
                total_dos_xdata = self.pdos_data["dos"][0]["x"]
                total_dos_ydata = self.pdos_data["dos"][0]["y"]
                x_min, x_max = get_x_min_max(fermi_energy)
                update_layout_non_spin_polarized(
                    total_dos_xdata,
                    total_dos_ydata,
                    x_min,
                    x_max,
                    update_func,
                    layout_type,
                )
            else:
                x_data_up = self.pdos_data["dos"][0]["x"]
                y_data_up = self.pdos_data["dos"][0]["y"]
                x_data_down = self.pdos_data["dos"][1]["x"]
                y_data_down = self.pdos_data["dos"][1]["y"]
                x_min, x_max = get_x_min_max(fermi_energy)
                update_layout_spin_polarized(
                    x_data_up,
                    y_data_up,
                    x_data_down,
                    y_data_down,
                    x_min,
                    x_max,
                    update_func,
                    layout_type,
                )

        # PDOS plot type
        if self.plot_type == "pdos":
            fermi_energy = self.fermi_energy.get(
                "fermi_energy", self.fermi_energy.get("fermi_energy_up")
            )
            handle_spin_polarization(fermi_energy, fig.update_layout, "layout")

        # Combined plot type
        if self.plot_type == "combined":
            fermi_energy = self.fermi_energy.get(
                "fermi_energy", self.fermi_energy.get("fermi_energy_up")
            )
            handle_spin_polarization(fermi_energy, fig.update_xaxes, "xaxes")



from .process import QeAppWorkChainSelector, WorkChainSelector
from .state import STATE_ICONS
from .tree import SimplifiedProcessTree, SimplifiedProcessTreeModel

__all__ = [
    "STATE_ICONS",
    "QeAppWorkChainSelector",
    "SimplifiedProcessTree",
    "SimplifiedProcessTreeModel",
    "WorkChainSelector",
]



"""Widgets for the submission of bands work chains.

Authors: AiiDAlab team
"""

from __future__ import annotations

import ipywidgets as ipw

from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.app.utils import get_entry_items
from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.panel import (
    ConfigurationSettingsModel,
    ConfigurationSettingsPanel,
)
from aiidalab_qe.common.widgets import LinkButton
from aiidalab_qe.common.wizard import QeConfirmableDependentWizardStep

from .advanced import (
    AdvancedConfigurationSettingsModel,
    AdvancedConfigurationSettingsPanel,
)
from .basic import BasicConfigurationSettingsModel, BasicConfigurationSettingsPanel
from .model import ConfigurationStepModel

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore


class ConfigureQeAppWorkChainStep(
    QeConfirmableDependentWizardStep[ConfigurationStepModel]
):
    missing_information_warning = "Missing input structure. Please set it first."

    def __init__(self, model: ConfigurationStepModel, **kwargs):
        super().__init__(
            model=model,
            confirm_kwargs={
                "tooltip": "Confirm the currently selected settings and go to the next step",
            },
            **kwargs,
        )
        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )

        workchain_model = BasicConfigurationSettingsModel()
        self.workchain_settings = BasicConfigurationSettingsPanel(model=workchain_model)
        self._model.add_model("workchain", workchain_model)

        advanced_model = AdvancedConfigurationSettingsModel()
        self.advanced_settings = AdvancedConfigurationSettingsPanel(
            model=advanced_model
        )
        self._model.add_model("advanced", advanced_model)

        # HACK due to spin orbit moving to basic settings (#984), we need to
        # sync the basic model's spin orbit when the advanced model's spin
        # orbit is set from loaded process
        ipw.dlink(
            (self._model.get_model("advanced"), "spin_orbit"),
            (self._model.get_model("workchain"), "spin_orbit"),
        )

        self.settings = {
            "workchain": self.workchain_settings,
            "advanced": self.advanced_settings,
        }

        self.installed_property_children = []
        self.not_installed_property_children = []
        self._fetch_plugin_calculation_settings()

    def _render(self):
        super()._render()

        self._fetch_not_installed_property()

        # RelaxType: degrees of freedom in geometry optimization
        self.relax_type_help = ipw.HTML()
        ipw.dlink(
            (self._model, "relax_type_help"),
            (self.relax_type_help, "value"),
        )
        self.relax_type = ipw.ToggleButtons()
        ipw.dlink(
            (self._model, "relax_type_options"),
            (self.relax_type, "options"),
        )
        ipw.link(
            (self._model, "relax_type"),
            (self.relax_type, "value"),
        )

        self.tabs = ipw.Tab(
            layout=ipw.Layout(min_height="250px"),
            selected_index=None,
        )
        self.tabs.observe(
            self._on_tab_change,
            "selected_index",
        )

        self.install_new_plugin_button = LinkButton(
            description="Plugin store",
            link="plugin_manager.ipynb",
            icon="puzzle-piece",  # More intuitive icon
            button_style="primary",  # Keeps it prominent
            tooltip="Browse and install additional plugins from the Plugin Store",
        )

        self.sub_steps = ipw.Accordion(
            children=[
                ipw.VBox(
                    children=[
                        InAppGuide(identifier="properties-selection"),
                        *self.installed_property_children,
                        ipw.HTML("<hr>"),
                        ipw.HTML(
                            value="""
                            <p style="font-size:14px; line-height:1.6;">
                                The following additional property calculations are available in the
                                <b>Plugin registry</b> but are currently disabled because the required plugins are not installed.
                            </p>
                            <p style="font-size:14px; line-height:1.6;">
                                To enable them, please visit the <b>Plugin store</b> and install the necessary plugins.
                            </p>
                            """,
                            layout=ipw.Layout(margin="10px 0px"),
                        ),
                        self.install_new_plugin_button,
                        *self.not_installed_property_children,
                    ]
                ),
                ipw.VBox(
                    children=[
                        InAppGuide(identifier="calculation-settings"),
                        self.tabs,
                    ],
                ),
            ],
            layout=ipw.Layout(margin="10px 2px"),
            selected_index=None,
        )
        self.sub_steps.set_title(0, "Step 2.1: Select which properties to calculate")
        self.sub_steps.set_title(1, "Step 2.2: Customize calculation parameters")

        self.content.children = [
            InAppGuide(identifier="configuration-step"),
            ipw.HTML("""
                <div style="padding-top: 0px; padding-bottom: 0px">
                    <h4>Structure relaxation</h4>
                </div>
            """),
            self.relax_type_help,
            self.relax_type,
            self.sub_steps,
        ]

        self.children = [
            self.content,
            self.confirm_box,
        ]

    def _post_render(self):
        self._update_tabs()

    def reset(self):
        self._model.reset()
        if self.rendered:
            self.sub_steps.selected_index = None
            self.tabs.selected_index = 0

    def _on_tab_change(self, change):
        if (tab_index := change["new"]) is None:
            return
        tab: ConfigurationSettingsPanel = self.tabs.children[tab_index]  # type: ignore
        tab.render()
        tab.update()

    def _on_input_structure_change(self, _):
        self._model.update()
        self.reset()

    def _update_tabs(self):
        children = []
        titles = []
        for identifier, model in self._model.get_models():
            if model.include:
                settings = self.settings[identifier]
                titles.append(model.title)
                children.append(settings)
        if self.rendered:
            self.tabs.selected_index = None
            self.tabs.children = children
            for i, title in enumerate(titles):
                self.tabs.set_title(i, title)
            self.tabs.selected_index = 0

    def _update_state(self, _=None):
        if self._model.confirmed:
            self.state = self.State.SUCCESS
        elif self.previous_step_state is self.State.SUCCESS:
            self.state = self.State.CONFIGURED
        elif self.previous_step_state is self.State.FAIL:
            self.state = self.State.FAIL
        else:
            self.state = self.State.INIT

    def _fetch_not_installed_property(self, plugin_config_source=None):
        from aiidalab_qe.app.utils.plugin_manager import (
            DEFAULT_PLUGIN_CONFIG_SOURCE,
            PluginManager,
            is_package_installed,
        )

        self.not_installed_property_children = []

        plugin_config_source = plugin_config_source or DEFAULT_PLUGIN_CONFIG_SOURCE
        plugin_manager = PluginManager(plugin_config_source)
        for plugin_name, plugin_data in plugin_manager.data.items():
            if (
                plugin_data.get("category", "calculation").lower() != "calculation"
            ):  # Ignore non-property plugins
                continue

            is_installed = is_package_installed(plugin_name)
            if not is_installed:
                checkbox = ipw.Checkbox(
                    value=False,
                    description=plugin_data["title"],
                    disabled=True,
                    indent=False,
                    style={"description_width": "initial"},
                )
                self.not_installed_property_children.append(checkbox)

    def _fetch_plugin_calculation_settings(self):
        outlines = get_entry_items("aiidalab_qe.properties", "outline")
        entries = get_entry_items("aiidalab_qe.properties", "configuration")
        for identifier, configuration in entries.items():
            for key in ("panel", "model"):
                if key not in configuration:
                    raise ValueError(f"Entry {identifier} is missing the '{key}' key")

            model: ConfigurationSettingsModel = configuration["model"]()
            self._model.add_model(identifier, model)

            outline = outlines[identifier]()
            info = ipw.HTML()
            ipw.link(
                (model, "include"),
                (outline.include, "value"),
            )

            if identifier == "bands":
                ipw.dlink(
                    (self._model, "input_structure"),
                    (outline.include, "disabled"),
                    lambda _: not self._model.has_pbc,
                )

            def toggle_plugin(change, identifier=identifier, model=model, info=info):
                if change["new"]:
                    info.value = (
                        f"Customize {identifier} settings in <b>Step 2.2</b> if needed"
                    )
                else:
                    info.value = ""
                model.update()
                self._update_tabs()

            model.observe(
                toggle_plugin,
                "include",
            )

            self.installed_property_children.append(
                ipw.HBox(
                    children=[
                        outline,
                        info,
                    ]
                )
            )

            panel: ConfigurationSettingsPanel = configuration["panel"](model=model)
            self.settings[identifier] = panel



from __future__ import annotations

import io

import ipywidgets as ipw
import traitlets as tl

from aiida import orm
from aiida.plugins import DataFactory, GroupFactory
from aiidalab_qe.common.widgets import HBoxWithUnits, LoadingWidget
from aiidalab_widgets_base.utils import StatusHTML

from ..subsettings import AdvancedConfigurationSubSettingsPanel
from .model import PseudosConfigurationSettingsModel

UpfData = DataFactory("pseudo.upf")
SsspFamily = GroupFactory("pseudo.family.sssp")
PseudoDojoFamily = GroupFactory("pseudo.family.pseudo_dojo")
CutoffsPseudoPotentialFamily = GroupFactory("pseudo.family.cutoffs")


class PseudosConfigurationSettingsPanel(
    AdvancedConfigurationSubSettingsPanel[PseudosConfigurationSettingsModel],
):
    def __init__(self, model: PseudosConfigurationSettingsModel, **kwargs):
        super().__init__(model, **kwargs)

        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )
        self._model.observe(
            self._on_protocol_change,
            "protocol",
        )
        self._model.observe(
            self._on_spin_orbit_change,
            "spin_orbit",
        )
        self._model.observe(
            self._on_family_parameters_change,
            ["library", "functional"],
        )
        self._model.observe(
            self._on_family_change,
            "family",
        )
        self._model.observe(
            self._on_reset_trigger_change,
            "pseudo_filename_reset_trigger",
        )

        ipw.dlink(
            (self._model, "cutoffs"),
            (self._model, "ecutwfc"),
            lambda cutoffs: max(cutoffs[0]),
        )
        ipw.dlink(
            (self._model, "cutoffs"),
            (self._model, "ecutrho"),
            lambda cutoffs: max(cutoffs[1]),
        )

    def render(self):
        if self.rendered:
            return

        self.family_prompt = ipw.HTML()

        self.family_help = ipw.HTML()
        ipw.dlink(
            (self._model, "family_help_message"),
            (self.family_help, "value"),
        )

        self.functional_help = ipw.HTML("""
            <div class="pseudo-text">
                The exchange-correlation energy is calculated using this functional.
                <br>
                We currently provide support for two well-established generalized
                gradient approximation (GGA) functionals: PBE and PBEsol.
            </div>
        """)

        self.functional = ipw.ToggleButtons()
        ipw.dlink(
            (self._model, "functional_options"),
            (self.functional, "options"),
        )
        ipw.link(
            (self._model, "functional"),
            (self.functional, "value"),
        )

        self.library = ipw.ToggleButtons()
        ipw.dlink(
            (self._model, "library_options"),
            (self.library, "options"),
        )
        ipw.link(
            (self._model, "library"),
            (self.library, "value"),
        )

        self.setter_widget = ipw.VBox()

        self._status_message = StatusHTML(clear_after=20)
        ipw.dlink(
            (self._model, "status_message"),
            (self._status_message, "message"),
        )

        self.ecutwfc = ipw.FloatText(
            description="Wavefunction",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "ecutwfc"),
            (self.ecutwfc, "value"),
        )
        self.ecutrho = ipw.FloatText(
            description="Charge density",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "ecutrho"),
            (self.ecutrho, "value"),
        )

        self.children = [
            ipw.HTML("<h2>Accuracy and precision</h2>"),
            ipw.HTML("""
                <div class="pseudo-text">
                    The exchange-correlation functional and pseudopotential library is
                    set by the <b>protocol</b> configured in the <b>Basic settings</b>
                    tab.
                    <br>
                    Here you can override the defaults if desired.
                </div>
            """),
            ipw.VBox(
                children=[
                    ipw.HTML("<h4>Exchange-correlation functional</h4>"),
                    self.functional,
                    self.functional_help,
                ],
            ),
            ipw.VBox(
                children=[
                    self.family_prompt,
                    self.library,
                    self.family_help,
                ],
            ),
            ipw.HTML("<h4>Pseudopotentials</h4>"),
            ipw.HTML("""
                <div class="pseudo-text">
                    The pseudopotential for each kind of atom in the structure can be
                    custom set.
                    <br>
                    The default pseudopotential and cutoffs are taken from the
                    pseudopotential family.
                    <br>
                    Recommended wavefunction (ψ) and charge density (ρ) cutoffs are
                    given to the right of each pseudopotential.
                </div>
            """),  # noqa: RUF001
            self.setter_widget,
            ipw.HTML("<h4>Cutoffs</h4>"),
            ipw.HTML("""
                <div style="line-height: 1.4;">
                    The
                    <a
                        href="https://www.quantum-espresso.org/Doc/INPUT_PW.html#idm312"
                        target="_blank"
                    >
                        default cutoffs
                    </a> used for the calculation are the maximum of the default cutoffs
                    from all pseudopotentials.
                    <br>
                    You can override them here.
                </div>
            """),
            HBoxWithUnits(self.ecutwfc, "Ry"),
            HBoxWithUnits(self.ecutrho, "Ry"),
            self._status_message,
        ]

        self.rendered = True

        self.refresh(specific="widgets")

    def _on_input_structure_change(self, _):
        self.refresh(specific="structure")

    def _on_protocol_change(self, _):
        self.refresh(specific="protocol")

    def _on_spin_orbit_change(self, _):
        self._model.update_library_options()

    def _on_family_parameters_change(self, _):
        self._model.update_family()

    def _on_family_change(self, _):
        self._update_family_link()
        self._model.update_default_pseudos()
        self._model.update_default_cutoffs()

    def _on_reset_trigger_change(self, _):
        if not self.rendered:
            return
        upload_widget: PseudoUploadWidget
        for upload_widget in self.setter_widget.children:
            upload_widget.reset_pseudo_filename_widget()

    def _update(self, specific=""):
        if self.updated:
            return
        self._show_loading()
        if not self._model.loaded_from_process or (specific and specific != "widgets"):
            self._model.update(specific)
        self._build_setter_widgets()
        self._model.update_library_options()
        self._update_family_link()
        self.updated = True

    def _update_family_link(self):
        if not self.rendered:
            return

        library, accuracy = self._model.library.split()
        if library == "SSSP":
            pseudo_family_link = (
                f"https://www.materialscloud.org/discover/sssp/table/{accuracy}"
            )
        else:
            pseudo_family_link = "http://www.pseudo-dojo.org/"

        self.family_prompt.value = f"""
            <h4>
                <a href="{pseudo_family_link}" target="_blank">
                    Pseudopotential family
                </a>
            </h4>
        """

    def _show_loading(self):
        if self.rendered:
            self.setter_widget.children = [self.loading_message]

    def _build_setter_widgets(self):
        if not self.rendered:
            return

        children = []

        kinds = self._model.input_structure.kinds if self._model.input_structure else []

        for index, kind in enumerate(kinds):
            upload_widget = PseudoUploadWidget(kind_name=kind.name)
            pseudo_link = ipw.link(
                (self._model, "dictionary"),
                (upload_widget, "pseudo"),
                [
                    lambda dictionary, kind_name=kind.name: orm.load_node(
                        dictionary.get(kind_name)
                    ),
                    lambda pseudo, kind_name=kind.name: {
                        **self._model.dictionary,
                        kind_name: pseudo.uuid,
                    },
                ],
            )
            cutoffs_link = ipw.dlink(
                (self._model, "cutoffs"),
                (upload_widget, "cutoffs"),
                lambda cutoffs, index=index: [cutoffs[0][index], cutoffs[1][index]]
                if len(cutoffs[0]) > index
                else [0.0, 0.0],
            )
            upload_widget.render()

            self.links.extend(
                [
                    pseudo_link,
                    cutoffs_link,
                    *upload_widget.links,
                ]
            )

            children.append(upload_widget)

        self.setter_widget.children = children


# TODO implement/improve MVC in this widget
class PseudoUploadWidget(ipw.HBox):
    """Class that allows to upload pseudopotential from user's computer."""

    pseudo = tl.Instance(UpfData, allow_none=True)
    cutoffs = tl.List(tl.Float(), [])
    error_message = tl.Unicode(allow_none=True)

    def __init__(self, kind_name, **kwargs):
        super().__init__(
            children=[LoadingWidget("Loading pseudopotential uploader")],
            **kwargs,
        )

        self.kind_name = kind_name

        self.rendered = False

    def render(self):
        if self.rendered:
            return

        self.pseudo_text = ipw.Text(
            description=self.kind_name,
            style={"description_width": "50px"},
        )
        pseudo_link = ipw.dlink(
            (self, "pseudo"),
            (self.pseudo_text, "value"),
            lambda pseudo: pseudo.filename if pseudo else "",
        )
        self.file_upload = ipw.FileUpload(
            description="Upload",
            multiple=False,
            layout=ipw.Layout(
                width="fit-content",
                margin="2px 10px 2px 2px",
            ),
        )
        self.file_upload.observe(self._on_file_upload, "value")

        cutoffs_message_template = """
            <div class="pseudo-text">
                ψ: <b>{ecutwfc} Ry</b> | ρ: <b>{ecutrho} Ry</b>
            </div>
        """  # noqa: RUF001

        self.cutoff_message = ipw.HTML()
        cutoff_link = ipw.dlink(
            (self, "cutoffs"),
            (self.cutoff_message, "value"),
            lambda cutoffs: cutoffs_message_template.format(
                ecutwfc=cutoffs[0] if len(cutoffs) else "not set",
                ecutrho=cutoffs[1] if len(cutoffs) else "not set",
            ),
        )

        self.links = [
            pseudo_link,
            cutoff_link,
        ]

        self.error_message = None

        self.children = [
            self.pseudo_text,
            self.file_upload,
            self.cutoff_message,
        ]

        self.rendered = True

    def reset_pseudo_filename_widget(self):
        self.pseudo_text.value = self.pseudo.filename if self.pseudo else ""

    def _on_file_upload(self, change=None):
        """When file upload button is pressed."""
        filename, item = next(iter(change["new"].items()))
        content = item["content"]

        # Order matters make sure when pseudo change
        # the pseudo_filename is set
        with self.hold_trait_notifications():
            self.pseudo = UpfData(io.BytesIO(content), filename=filename)
            self.pseudo.store()

            # check if element is matched with the pseudo
            element = "".join([i for i in self.kind_name if not i.isdigit()])
            if element != self.pseudo.element:
                self.error_message = f"""<div class="alert alert-danger"> ERROR: Element {self.kind_name} is not matched with the pseudo {self.pseudo.element}</div>"""
                self._reset()
            else:
                self.pseudo_text.value = filename

    def _reset(self):
        """Reset the widget to the initial state."""
        self.pseudo = None
        self.cutoffs = []



from aiidalab_qe.app.wrapper import AppWrapperContoller, AppWrapperModel, AppWrapperView


class TestWrapper:
    def test_enable_controls(self):
        """Test enable_controls method."""
        self._instansiate_mvc_components()
        assert self.view.guide_toggle.disabled is True
        assert self.view.about_toggle.disabled is True
        self.controller.enable_controls()
        assert self.view.guide_toggle.disabled is False
        assert self.view.about_toggle.disabled is False

    def test_guide_toggle(self):
        """Test guide_toggle method."""
        self._instansiate_mvc_components()
        self.controller.enable_controls()
        self.controller._on_guide_toggle({"new": True})
        self._assert_guide_is_on()
        self.controller._on_guide_toggle({"new": False})
        self._assert_no_info()

    def test_about_toggle(self):
        """Test about_toggle method."""
        self._instansiate_mvc_components()
        self.controller.enable_controls()
        self.controller._on_about_toggle({"new": True})
        self._assert_about_is_on()
        self.controller._on_about_toggle({"new": False})
        self._assert_no_info()

    def test_toggle_switch(self):
        """Test toggle_switch method."""
        self._instansiate_mvc_components()
        self.controller.enable_controls()
        self._assert_no_info()
        self.controller._on_guide_toggle({"new": True})
        self._assert_guide_is_on()
        self.controller._on_about_toggle({"new": True})
        self._assert_about_is_on()
        self.controller._on_guide_toggle({"new": True})
        self._assert_guide_is_on()
        self.controller._on_guide_toggle({"new": False})
        self._assert_no_info()

    def _assert_guide_is_on(self):
        """Assert guide is on."""
        assert len(self.view.info_container.children) == 2
        assert self.view.guide in self.view.info_container.children

    def _assert_about_is_on(self):
        """Assert about is on."""
        assert len(self.view.info_container.children) == 1
        assert self.view.about in self.view.info_container.children

    def _assert_no_info(self):
        """Assert no info is shown."""
        assert len(self.view.info_container.children) == 0

    def _instansiate_mvc_components(self):
        """Instansiate `AppWrapper` MVC components."""
        self.model = AppWrapperModel()
        self.view = AppWrapperView()
        self.controller = AppWrapperContoller(self.model, self.view)



from __future__ import annotations

import os
from collections.abc import Iterable
from dataclasses import dataclass, field
from pathlib import Path
from subprocess import run

from aiida_pseudo.groups.family import PseudoPotentialFamily
from filelock import FileLock, Timeout

from aiida.orm import QueryBuilder

SSSP_VERSION = "1.3"
PSEUDODOJO_VERSION = "0.4"

EXPECTED_PSEUDOS = {
    f"SSSP/{SSSP_VERSION}/PBE/efficiency",
    f"SSSP/{SSSP_VERSION}/PBE/precision",
    f"SSSP/{SSSP_VERSION}/PBEsol/efficiency",
    f"SSSP/{SSSP_VERSION}/PBEsol/precision",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBE/SR/standard/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBEsol/SR/standard/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBE/SR/stringent/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBEsol/SR/stringent/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBE/FR/standard/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBEsol/FR/standard/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBE/FR/stringent/upf",
    f"PseudoDojo/{PSEUDODOJO_VERSION}/PBEsol/FR/stringent/upf",
}


FN_LOCKFILE = Path.home().joinpath(".install-sssp.lock")


@dataclass
class PseudoFamily:
    """The dataclass to deal with pseudo family strings.

    Attributes:
    library: the library name of the pseudo family, e.g. SSSP or PseudoDojo.
    cmd_library_name: the sub command name used in aiida-pseudo command line.
    version: the version of the pseudo family, e.g. 1.2
    functional: the functional of the pseudo family, e.g. PBE, PBEsol.
    accuracy: the accuracy of the pseudo family, which is protocol in aiida-pseudo, e.g. efficiency, precision, standard, stringent.
    relativistic: the relativistic treatment of the pseudo family, e.g. SR, FR.
    file_type: the file type of the pseudo family, e.g. upf, psml, currently only used for PseudoDojo.
    """

    library: str
    version: str
    functional: str
    accuracy: str
    cmd_library_name: str = field(init=False)
    relativistic: str | None = None
    file_type: str | None = None

    def __post_init__(self):
        """Post init operations and checks."""
        if self.library == "SSSP":
            self.cmd_library_name = "sssp"
        elif self.library == "PseudoDojo":
            self.cmd_library_name = "pseudo-dojo"
        else:
            raise ValueError(f"Unknown pseudo library {self.library}")

    @classmethod
    def from_string(cls, pseudo_family_string: str) -> PseudoFamily:
        """Initialize from a pseudo family string."""
        # We support two pseudo families: SSSP and PseudoDojo
        # They are formatted as follows:
        # SSSP: SSSP/<version>/<functional>/<accuracy>
        # PseudoDojo: PseudoDojo/<version>/<functional>/<relativistic>/<accuracy>/<file_type>
        # where <relativistic> is either 'SR' or 'FR' and <file_type> is either 'upf' or 'psml'
        # Before we unify the format of family strings, the conditions below are necessary
        # to distinguish between the two families
        library = pseudo_family_string.split("/")[0]
        if library == "SSSP":
            version, functional, accuracy = pseudo_family_string.split("/")[1:]
            relativistic = None
            file_type = None
        elif library == "PseudoDojo":
            (
                version,
                functional,
                relativistic,
                accuracy,
                file_type,
            ) = pseudo_family_string.split("/")[1:]
        else:
            raise ValueError(
                f"Not able to parse valid library name from {pseudo_family_string}"
            )

        return cls(
            library=library,
            version=version,
            functional=functional,
            accuracy=accuracy,
            relativistic=relativistic,
            file_type=file_type,
        )


def pseudos_to_install() -> set[str]:
    """Query the database and return the list of pseudopotentials that are not installed."""
    qb = QueryBuilder()
    qb.append(
        PseudoPotentialFamily,
        filters={
            "or": [
                {"label": {"like": "SSSP/%"}},
                {"label": {"like": "PseudoDojo/%"}},
            ]
        },
        project="label",
    )
    labels = set(qb.all(flat=True))
    return EXPECTED_PSEUDOS - labels


def _construct_cmd(
    pseudo_family_string: str, download_only: bool = False, cwd: Path | None = None
) -> list:
    """Construct the command for installation of pseudopotentials.

    If ``cwd`` is not None, and ``download_only`` is True the, only download the
    pseudopotential files to the ``cwd`` folder.
    If ``download_only`` is False and ``cwd`` is not None, the the pseudos will be installed from the ``cwd`` where the pseudos are downloaded to.

    NOTE: download_only has nothing to do with cwd, it will not download the pseudos to cwd if cwd is specified.
    The control to download to cwd is in the ``_install_pseudos`` function below.
    """
    pseudo_family = PseudoFamily.from_string(pseudo_family_string)

    # the library used in command line is lowercase
    # e.g. SSSP -> sssp and PseudoDojo -> pseudo-dojo
    library = pseudo_family.cmd_library_name
    version = pseudo_family.version
    functional = pseudo_family.functional
    accuracy = pseudo_family.accuracy
    cmd = [
        "aiida-pseudo",
        "install",
        library,
        "--functional",
        functional,
        "--version",
        version,
        "-p",  # p for protocol which is the accuracy of the library
        accuracy,
    ]

    # extra arguments for PseudoDojo
    if library == "pseudo-dojo":
        relativistic = pseudo_family.relativistic
        file_type = pseudo_family.file_type
        cmd.extend(
            [
                "--relativistic",
                relativistic,
                "--pseudo-format",
                file_type,
            ]
        )

    if download_only:
        cmd.append("--download-only")

    # if cwd source folder specified, then install the pseudos from the folder
    # download file name is replace `/` with `_` of the pseudo family string with `.aiida_pseudo` extension
    if not download_only and cwd is not None:
        file_path = cwd / f"{pseudo_family_string.replace('/', '_')}.aiida_pseudo"
        if file_path.exists():
            cmd.extend(["--from-download", str(file_path)])

    return cmd


def run_cmd(cmd: list, env: dict | None = None, cwd: Path | None = None):
    """Run the command with specific env in the workdir specified."""
    run(cmd, env=env, cwd=cwd, capture_output=True, check=True)


def _install_pseudos(
    pseudo_families: set[str], download_only: bool = False, cwd: Path | None = None
) -> Iterable[float]:
    """Go through the list of pseudo families and install them."""
    env = os.environ.copy()
    env["PATH"] = f"{env['PATH']}:{Path.home() / '.local' / 'bin'}"

    mult = 1.0 / len(pseudo_families)
    yield mult * 0
    for i, pseudo_family in enumerate(pseudo_families):
        cmd = _construct_cmd(pseudo_family, download_only, cwd=cwd)

        run_cmd(cmd, env=env, cwd=cwd)

        yield mult * (i + 1)


def install(
    download_only: bool = False, cwd: Path | None = None
) -> Iterable[tuple[str, float]]:
    yield "Checking installation status...", 0.1
    try:
        with FileLock(FN_LOCKFILE, timeout=5):
            if len(pseudos := pseudos_to_install()) > 0:
                yield "Installing...", 0.1
                for progress in _install_pseudos(pseudos, download_only, cwd):
                    yield "Installing...", progress

    except Timeout:
        # Assume that the installation was triggered by a different process.
        from aiidalab_qe.common.widgets import ProgressBar

        yield (
            "Installation was already started elsewhere, waiting for it to finish...",
            ProgressBar.AnimationRate(1.0),
        )
        with FileLock(FN_LOCKFILE, timeout=300):
            if len(pseudos_to_install()) > 0:
                raise RuntimeError(
                    "Installation process did not finish in the expected time."
                ) from None
